#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''
Phase *history* container. When you think of a phase value, think of that value at
a specific point in time. This container holds the historic data of a phase;
its species and quantities. This implementation treats access of time stamps within
a tolerance. All searches for time stamped values are subjected to an approximation
of the time stamp to avoid storing values too close to each other in time, and/or to
return the closest value in time searched or no value if none can be found according
to the tolerance.

Background
----------
TODO: ATTENTION:
The species (list of Species) AND quantities (list of Quantity) data members
have ARBITRARY density values either at an arbitrary point in the history or at
no point in the history. This needs to be removed in the future to avoid confusion.

To obtain history values, associated to the phase, at a particular point in time,
use the GetValue() method to access the history data frame (dict(list())) via columns
(dict) and rows (list).
ALERT: The corresponding values in species and quantities are OVERRIDEN and NOT to
be used through the phase interface.

Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda
Sat Sep  5 01:26:53 EDT 2015

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
import os, io
from copy import deepcopy
import time
import datetime

import numpy as np

import matplotlib
matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import MultipleLocator
from matplotlib.ticker import ScalarFormatter

from cortix.support.species   import Species
from cortix.support.quantity import Quantity

class PhaseNewest:
    '''
    Phase `history` container. A `Phase` consists of `Species` and `Quantities`
    varying with time. This container is meant to reproduce the basic idea of a
    material phase.
    '''

    def __init__(self, name = None,
                 time_stamp = None,
                 time_unit  = None,
                 species    = None,
                 quantities = None
                ):
        #TODO
        '''
        Sometimes an empty Phase object is created by user code. This case needs
        adequate logic for None types.
        Note on usage: when passing quantities, do set the value argument explicitly
        to help define the type and avoid set_value() errors with Pandas. This is
        to be investigated later. Also, the usage of a DataFrame needs to be re-evaluated.
        Maybe better to use a Quantity object and a Species object with a Pandas Series
        history as a value to avoid the existance of a value in Quantity and a value
        in Phase that are not in sync.
        '''

        if not name:
            self.__name = self.__class__.__name__
        else:
            self.__name = name

        if time_stamp is None:
            time_stamp = 0.0 # default type is float
        else:
            assert isinstance(time_stamp, float)

        if time_unit is None:
            self.__time_unit = 's' # second
        else:
            assert isinstance(time_unit, str)
            self.__time_unit = time_unit

        if species is not None:
            assert isinstance(species, list)
            for each_species in species:
                assert isinstance(each_species, Species)

        if quantities is not None:
            assert isinstance(quantities, list)
            for quant in quantities:
                assert isinstance(quant, Quantity)

        # List of species and quantities objects; columns of data frame are named
        # by objects.
        # A new object held by a Phase() object
        self.__species = deepcopy(species)
        # A new object held by a Phase() object
        self.__quantities = deepcopy(quantities)

        self.__names = list() # ordered list of names

        # Table data phase without data type assigned; this is left to the user
        # Time stamps will always be float
        #self.__df = pandas.DataFrame( index=[float(time_stamp)], columns=names )
        self.__df = dict(list())

        if species is not None:
            for spc in self.__species:
                name = spc.name
                assert name not in self.__names
                self.__names.append(name)

        if quantities is not None:
            for quant in self.__quantities:
                name = quant.name
                assert name not in self.__names
                self.__names.append(name)
                quant.value = 0.0    # clear these values
                                     # todo: eliminate them from Quantity in the future

        self.__df['time-stamps'] = [time_stamp]

        # This is meant to be the value of species concentration; a float type
        if species is not None:
            for spc in species:
                self.__df.loc[spc.name] = [0.0]

        if quantities is not None:
            for quant in quantities:  # attention here: using the original data
                self.__df[quant.name] = [quant.value]

        return

    def has_time_stamp(self, try_time_stamp):
        '''
        Checks to see if try_time_stamp exists in the phase history.

        Parameters
        ----------
        try_time_stamp:
        '''

        time_stamp = self.__get_time_stamp( try_time_stamp )

        if time_stamp is not None:
            return True
        else:
            return False

    def __get_time_unit(self):
        '''
        Returns the time unit of the `Phase.`

        Returns
        -------
        time_unit: str
        '''

        return self.__time_unit
    time_unit = property(__get_time_unit,None,None,None)

    def __get_time_stamps(self):
        '''
        Get all time stamps in the index of the data frame.

        Returns
        -------
        time_stamps: list
        '''

        #return list(self.__df.index)  # return all time stamps
        return self.__df.['time-stamps']  # return all time stamps
    time_stamps = property(__get_time_stamps, None, None, None)

    def __get_species_list(self):
        '''
        Returns every single species in the phase history.

        Returns
        -------
        species: list
        '''

        return self.__species
    species = property(__get_species_list, None, None, None)

    def GetQuantities(self):
        '''
        Returns the list of `Quantities`. The values in each `Quantity` are
        synchronized with the `Phase` data frame.

        Returns
        -------
        quantities: list
        '''

        for quant in self.__quantities:
          tmp = self.GetQuantity(quant.name) # handy way to synchronize the whole list
        return self.__quantities
    quantities = property(GetQuantities, None, None, None)

    def __get_actors(self):
        '''
        Returns a list of names of all the actors in the phase history.

        Returns
        -------
        #list(self.__df.colums): list
        names: list

        '''

        #return list(self.__df.columns)  # return all names in order
        return self.__names # return all names in order
    actors = property(__get_actors, None, None, None)

    def __get_df(self):
        '''
        Die hard access.
        '''
        return self.__df
    df = property(__get_df,None,None,None)

    def get_species(self, name):
        '''
        Returns the species specified by name if it exists,
        or None if it doesn't.

        Parameters
        ----------
        name: str

        Returns
        -------
        specie: str

        '''
        for species in self.__species:
            if species.name == name:
                return species
        return None

    def get_species_concentration(self, name, try_time_stamp=None):
        '''
        Returns the species concentration at `try_time_stamp`.

        Parameters
        ----------
        name: str

        try_time_time_stamp: float

        Returns
        -------
        concentration: float

        '''
        return self.get_value(name,try_time_stamp)

    def set_species_id(self, name, val):
        '''
        Sets the flag of a species "name" equal to val.

        Parameters
        ----------
        name: str
        val: int

        '''
        for species in self.__species:
            if species.name == name:
                species.flag = val
                return

    def get_quantity(self, name, try_time_stamp=None):
        '''
        Get the quantity `name` at a point in time closest to
        `try_time_stamp` up to a tolerance. If no time stamp is passed,
        the value at the last time stamp is returned.

        Parameters
        ----------
        name: str
        try_time_stamp: float, int or None
            Time stamp of desired quantity value. Default: None, returns the
            value at the last time stamp.

        Returns
        -------
        quant.value: float or int or other
        '''

        assert name in self.__df.columns, 'name %r not in %r'%\
                (name,self.__df.columns)

        time_stamp = self.__get_time_stamp( try_time_stamp )

        for quant in self.__quantities:
            if quant.name == name:
                quant.value = self.__df.loc[time_stamp, name] # labels' access mode
                return quant  # return quantity syncronized with the phase

    def get_quantity_history(self, name):
        '''
        Create a Quantity `name` history. This will create a fully qualified
        Quantity object and return to the caller. The function is typically
        needed for data output to a file through `pickle`. Since the value
        attribute of a quantity can be any data structure, a time-series is
        built on the fly and stored in the value attribute. In addition the
        time unit is added to the final return value as a tuple.

        Parameters
        ----------
        name: str

        Returns
        -------
        quant_history: tuple(Quantity,str)
        '''

        assert name in self.__df.columns, 'name %r not in %r'%\
                (name,self.__df.columns)

        for quant in self.__quantities:
            if quant.name == name:
                quant_history = deepcopy(quant)
                quant_history.value = deepcopy(self.__df[name]) # whole data frame index series
                return (quant_history,self.__time_unit) # return tuple

    def add_single_species(self, new_species):
        '''
        Adds a new specie object to the phase history. See species.py for
        more details on the Species class.

        Parameters
        ----------
        new_species: obj
        '''

        assert isinstance(new_species, Species)
        assert new_species.name not in list(self.__df.columns), \
               'new_species: %r exists. Current names: %r' % \
               (new_species, self.__df.columns)
        species_formulae = [specie.formula_name for specie in self.__species]
        assert new_species.formula_name not in species_formulae
        self.__species.append(new_species)
        new_name = new_species.name
        col = pandas.DataFrame( index=list(self.__df.index), columns=[new_name] )
        tmp = self.__df
        df = tmp.join(col, how='outer')
        self.__df = df.fillna(0.0)   # for species have float as default

    def add_quantity(self, new_quant):
        '''
        Adds a new quantity object to the dataframe. See quantity.py for more
        details on the quantity class.

        Parameters
        ----------
        new_quant: object
        '''

        assert isinstance(new_quant, Quantity)
        assert new_quant.name not in list(self.__df.columns), \
               'quantity: %r exists. Current names: %r' % \
               (new_quant, self.__df.columns)
        quant_formal_names = [quant.formal_name for quant in self.__quantities]
        assert new_quant.formal_name not in quant_formal_names
        self.__quantities.append(new_quant)
        new_name = new_quant.name

        # create a col with object data type; user must fill out column
        col = pandas.DataFrame( index=list( self.__df.index), columns=[new_name],
                                dtype=object )
        tmp = self.__df
        df  = tmp.join(col, how='outer')
        #self.__df = df.fillna(new_quant.value)

    def add_row(self, try_time_stamp, row_values):
        '''
        Adds a row to the `DataFrame`, with a `timestamp` equal to `try_time_stamp` and
        row values equal to `row_values`. The length of `row_values` must match the
        number of columns in the data frame.

        Parameters
        ----------
        try_time_stamp: float
        row_values: list

        '''
        assert try_time_stamp not in self.__df.index, 'already used time_stamp: %r'%\
                (try_time_stamp)
        assert isinstance(row_values, list)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is None, 'already used time_stamp: %r'%(try_time_stamp)
        time_stamp = try_time_stamp

        assert len(row_values) == self.__df.columns.size

        # create a row with object data type; users row_values data define data type
        row = pandas.DataFrame( index=[time_stamp],
                                columns=list( self.__df.columns ), dtype=object )

        for (col,v) in zip(row.columns, row_values):
            row.loc[time_stamp,col] = v

        frames = [self.__df, row]
        self.__df = pandas.concat(frames)
        return

    def get_row(self, try_time_stamp=None):
        '''
        Returns an entire row of the phase dataframe. A row is a series of
        values that are all at the same time stamp.

        Parameters
        ----------
        try_time_stamp: float

        Returns
        -------
        list(self.__df.loc[time_stamp, :]): list

        '''
        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        return list(self.__df.loc[time_stamp, :])

    def get_column(self, actor):
        '''
        Returns an entire column of data. A column is the entire history
        of data associated with a specific actor.

        Parameters
        ----------
        actor: str

        Returns
        -------
        list(self.__df.loc[:, actor]): list

        '''
        assert isinstance(actor, str)
        assert actor in self.__df.columns, 'actor %r not in %r'% \
                   (actor,self.__df.columns)
        return list(self.__df.loc[:, actor])

    def scale_row(self, try_time_stamp, value):
        '''
        Multiplies all of the data in a row (except time stamp) by a scalar
        value.

        Parameters
        ----------
        try_time_stamp: float
        value: float

        '''
        assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)
        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        #assert isinstance(value, int) or isinstance(value, float)
        self.__df.loc[time_stamp, :] *= value
        return

    def ClearHistory(self, value=0.0):
        '''
        Set species and quantities of history to a given value
        (default to zero value), all time stamps are preserved.

        Parameters
        ----------
        value: float

        '''
        assert isinstance(value, int) or isinstance(value, float)
        self.__df.loc[:, :] = value

        return

    def ResetHistory(self, try_time_stamp=None, value=None):
        '''
        Set species and quantities of history to a given value
        (default to zero value) only one time stamp is preserved (default to
        last time stamp).

        Parameters
        ----------
        try_time_stamp: float
        value: float

        '''
        if value is not None:
           assert isinstance(value, int) or isinstance(value, float) or \
                  isinstance(value, np.ndarray)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        values = self.GetRow(time_stamp)  # save values

        columns = list(self.__df.columns)
        assert len(columns) == len(values), 'FATAL: oops internal error.'

        self.__df = pandas.DataFrame( index=[time_stamp], columns=columns )
        self.__df.fillna( 0.0, inplace=True )

        if value is None:
            for v in values:
                idx = values.index(v)
                self.__df.loc[time_stamp, columns[idx]] = v  # restore values
        else:
            self.__df.loc[time_stamp, :] = value   # set user-given value

        return

    def get_value(self, actor, try_time_stamp=None):
        '''
        Returns the value associated with a specified actor at a specified
        time stamp.

        Parameters
        ----------
        actor: str
        try_time_stamp: float
            Default is None which returns the last time stamp.

        Returns
        -------
        self.__df.loc[time_stamp, actor]: any

        '''
        assert isinstance(actor, str)
        assert actor in self.__df.columns, 'actor %r not in %r'% \
                   (actor,self.__df.columns)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        return self.__df.loc[time_stamp, actor]

    def set_value(self, actor, value, try_time_stamp=None):
        '''

        '''
        assert isinstance(actor, str)
        assert actor in self.__df.columns

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        # Note: user value could have a different type than other column values.
        # If there is a type change, this will not be checked; user has been advised.
        self.__df.loc[time_stamp, actor] = value

        return

    def write_html(self, fileName):
        '''
        Convert the `Phase` container into an HTML file.

        Parameters
        ---------
        fileName: str

        '''
        assert isinstance(fileName, str)
        tmp = pandas.DataFrame(self.__df)
        column_names = tmp.columns
        species_names = [species.name for species in self.__species]
        quantity_names = [quantity.name for quantity in self.__quantities]
        for col in column_names:
            if col in species_names:
                idx = species_names.index(col)
                species = self.__species[idx]
                tmp.rename(columns={col: species.formula_name}, inplace=True)
            elif col in quantity_names:
                idx = quantity_names.index(col)
                quant = self.__quantities[idx]
                tmp.rename( columns={ col: col + '[' + quant.unit + ']'},
                            inplace=True )
            else:
                assert False, 'oops fatal.'

        tmp.to_html(fileName)

        return

    def __str__(self):
        s = '\n\t **Phase()**: name=%s;' + \
            '\n\t time unit: %s;' + \
            '\n\t *quantities*: %s;' + \
            '\n\t *species*: %s;' + \
            '\n\t *history* # time_stamps=%s;' + \
            '\n\t *history end* @%s;' + \
            '\n%s'
        return s % (self.name,
                self.__time_unit,
                self.__quantities,
                self.__species,
                len(self.__df.index),
                self.__df.index[-1],
                self.__df.loc[self.__df.index[-1], :] )

    def __repr__(self):
        s = '\n\t **Phase()**: name=%s;' + \
            '\n\t time unit: %s;' + \
            '\n\t *quantities*: %s;' + \
            '\n\t *species*: %s;' + \
            '\n\t *history* # time_stamps=%s;' + \
            '\n\t *history end* @%s;' + \
            '\n%s'
        return s % (self.name,
                self.__time_unit,
                self.__quantities,
                self.__species,
                len(self.__df.index),
                self.__df.index[-1],
                self.__df.loc[self.__df.index[-1], :] )

    def __get_time_stamp(self, try_time_stamp=None):
        '''
        Helper method for finding the closest time stamp to `try_time_stamp`
        in the phase history. The pandas index container used for storing
        float data type time stamps will return the nearest time stamp up to a
        tolerance. Whether the time index has one value, this function will
        inspect for the proximity to that value.

        Parameters
        ----------
        try_time_stamp: float, int or None
            Default: None will return the last time stamp.

        Returns
        -------
        self.__df.index[loc]: float or None
            Will return None if no time stamp within tolerance is found.

        '''
        import numpy as np

        tol = 1.0e-3

        if try_time_stamp is None:
            return self.__df.index[-1]
        else:
            time_stamps = np.array(self.__df.index)
            if time_stamps.size >= 2:
               tol = 1.0e-3 * np.diff(time_stamps).mean() # 1e-3 * the mean delta t
            try: # abs(index_value - try_time_stamp) <= tolerance
                loc = self.__df.index.get_loc( try_time_stamp, method='nearest',
                        tolerance=tol )
            except KeyError: # no value found withing tol
                return None
            else:
                return  self.__df.index[loc]

    def plot_species(self, name, scaling=[1.0,1.0] , title=None, xlabel='Time [s]', 
            ylabel='y', legend='no-legend', filename_tag=None, figsize=[6,5], dpi=100 ):

        fig,ax=plt.subplots(1,figsize=figsize)

        x = np.array( [t for t in self.__df.index] )
        x *= float(scaling[0])

        y = np.array( self.get_column(name),dtype=np.float64 )
        y *= float(scaling[1])

        yformatter = ScalarFormatter(useMathText=True,useOffset=True)
        yformatter.set_powerlimits((15, 5))
        ax.yaxis.set_major_formatter(yformatter)

        ax.plot(x, y, 'b-', label=legend)

        ax.set_xlabel(r''+xlabel,fontsize=16)
        ax.set_ylabel(r''+ylabel,fontsize=16,color='black')
        ax.tick_params(axis='y',labelsize=14)
        ax.tick_params(axis='x',labelsize=14)
        ax.legend(loc='best',fontsize=12)

        if title:
            ax.set_title(title)
        elif self.get_species(name).info:
            ax.set_title(self.get_species(name).info,fontsize=14)

        ax.grid(True)

        fig_name = name+'-'+self.name+'-phase-plot-'

        if filename_tag:
            fig_name += filename_tag

        fig.savefig(fig_name+'.png', dpi=dpi, fomat='png')
        plt.close(fig)

        return

    def plot( self, name='phase-plot-name', time_unit='s', legend=None, 
            nrows=2, ncols=2, dpi=200):

        num_var = len(self.__df.columns)
        if num_var == 0:
            return

        today = datetime.datetime.today().strftime("%d%b%y %H:%M:%S")

        lead_name = name

        fig_num = None

        # Loop over variables and assign to the dashboards
        i_dash = 0
        for i_var in range(num_var):
            # if multiple of nrows*ncols start new dashboard
            if i_var % (nrows*ncols) == 0:

                if i_var != 0:  # flush any current figure
                    fig_name = lead_name+'-'+self.name+'-phase-plot-' + \
                            str(i_dash).zfill(2)
                    fig.savefig(fig_name+'.png', dpi=dpi, fomat='png')
                    plt.close(fig_num)

                    #pickle.dump( fig, open(fig_name+'.pickle','wb') )

                    i_dash += 1

                fig_num = str(np.random.random()) + '.' + str(i_dash)
                fig = plt.figure(num=fig_num)

                gs = gridspec.GridSpec(nrows, ncols)
#                gs.update(left=0.08, right=0.98, wspace=0.4, hspace=0.4)
                gs.update(left=0.11, right=0.98, wspace=0.4, hspace=0.5)

                axlst = list()

                nPlotsNeeded = num_var - i_var
                count = 0
                for i in range(nrows):
                    for j in range(ncols):
                        axlst.append(fig.add_subplot(gs[i, j]))
                        count += 1
                        if count == nPlotsNeeded:
                            break
                    if count == nPlotsNeeded:
                        break

                axes = np.array(axlst)

                text = today + ': Cortix.Phase.Plot'
                fig.text(.5, .95, text, horizontalalignment='center', fontsize=14)

                axs = axes.flat

                axId = 0

            # end of: if i_var % nrows*ncols == 0: # if a multiple of nrows*ncols
            # start a new dashboard

            ax = axs[axId]
            axId += 1

            col_name = self.__df.columns[i_var]

            species = self.get_species(col_name)
            if species:
                varName = species.formula_name
            else:
                quant = self.get_quantity(col_name)
                varName = quant.formal_name

            # sanity check
            if i_var <= len(self.__species):
                assert self.__species[i_var].name == self.__df.columns[i_var]
            else:
                assert self.__quantities[i_var].name == self.__df.columns[i_var]

            varUnit = 'g/L'

            '''
            if varUnit == 'gram':
                varUnit = 'g'
            if varUnit == 'gram/min':
                varUnit = 'g/min'
            if varUnit == 'gram/s':
                varUnit = 'g/s'
            if varUnit == 'gram/m3':
                varUnit = 'g/m3'
            if varUnit == 'gram/L':
                varUnit = 'g/L'
            if varUnit == 'sec':
                varUnit = 's'
            '''

            varLegend = legend
            varScale  = 'linear-linear'

            assert varScale == 'log' or varScale == 'linear' or varScale == 'log-linear' \
                or varScale == 'linear-log' or varScale == 'linear-linear' or \
                varScale == 'log-log'

            time_unit = 's'

            if time_unit == 'minute':
                time_unit = 'min'

            x = np.array( [i for i in self.__df.index] )

            if (varScale == 'linear' or varScale == 'linear-linear' or \
                varScale == 'linear-log') and x.max() >= 60.0:
                x /= 60.0
                if time_unit == 'min':
                    time_unit = 'h'
                if time_unit == 'second' or time_unit=='s':
                    time_unit = 'min'

            y = np.array( self.__df[col_name] )  # convert to numpy ndarray

            '''
            if (y.max() >= 1e3 or y.min() <= -1e3) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y /= 1e3
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'kg'
                if varUnit == 'L':
                    varUnit = 'kL'
                if varUnit == 'cc':
                    varUnit = 'L'
                if varUnit == 'Ci':
                    varUnit = 'kCi'
                if varUnit == 'W':
                    varUnit = 'kW'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'kg/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'kg/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'kg/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'kg/L'
                if varUnit == 'W/L':
                    varUnit = 'kW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'kCi/L'
                if varUnit == '':
                    varUnit = 'x1e3'
                if varUnit == 'L/min':
                    varUnit = 'kL/min'
                if varUnit == 'Pa':
                    varUnit = 'kPa'
                if varUnit == 's':
                    varUnit = 'ks'
                if varUnit == 'm':
                    varUnit = 'km'
                if varUnit == 'm/s':
                    varUnit = 'km/s'

            if (y.max() < 1e-6 and y.min() > -1e-6) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e9
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'ng'
                if varUnit == 'cc':
                    varUnit = 'n-cc'
                if varUnit == 'L':
                    varUnit = 'nL'
                if varUnit == 'W':
                    varUnit = 'nW'
                if varUnit == 'Ci':
                    varUnit = 'nCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'ng/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'ng/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'ng/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'ng/L'
                if varUnit == 'W/L':
                    varUnit = 'nW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'nCi/L'
                if varUnit == 'L/min':
                    varUnit = 'nL/min'
                if varUnit == 'Pa':
                    varUnit = 'nPa'
                if varUnit == 's':
                    varUnit = 'ns'
                if varUnit == 'm/s':
                    varUnit = 'nm/s'

            if (y.max() >= 1e-6 and y.max()  <  1e-3) or \
               (y.min() > -1e-3 and y.min() <= -1e-6) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e6
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'ug'
                if varUnit == 'cc':
                    varUnit = 'u-cc'
                if varUnit == 'L':
                    varUnit = 'uL'
                if varUnit == 'W':
                    varUnit = 'uW'
                if varUnit == 'Ci':
                    varUnit = 'uCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'ug/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'ug/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'ug/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'ug/L'
                if varUnit == 'W/L':
                    varUnit = 'uW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'uCi/L'
                if varUnit == 'L/min':
                    varUnit = 'uL/min'
                if varUnit == 'Pa':
                    varUnit = 'uPa'
                if varUnit == 's':
                    varUnit = 'us'
                if varUnit == 'm/s':
                    varUnit = 'um/s'

            if (y.max() >= 1e-3 and y.max()  < 1e-1) or \
               (y.min() <= -1e-3 and y.min() > -1e-1) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e3
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'mg'
                if varUnit == 'cc':
                    varUnit = 'm-cc'
                if varUnit == 'L':
                    varUnit = 'mL'
                if varUnit == 'W':
                    varUnit = 'mW'
                if varUnit == 'Ci':
                    varUnit = 'mCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'mg/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'mg/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'mg/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'mg/L'
                if varUnit == 'W/L':
                    varUnit = 'mW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'mCi/L'
                if varUnit == 'L/min':
                    varUnit = 'mL/min'
                if varUnit == 'Pa':
                    varUnit = 'mPa'
                if varUnit == 's':
                    varUnit = 'ms'
                if varUnit == 'm/s':
                    varUnit = 'mm/s'
            '''

            ax.set_xlabel('Time [' + time_unit + ']', fontsize=9)
            ax.set_ylabel(varName + ' [' + varUnit + ']', fontsize=9)

            '''
            ymax = y.max()
            dy = ymax * .1
            ymax += dy
            ymin = y.min()
            ymin -= dy

            if abs(ymin - ymax) <= 1.e-4:
                ymin = -1.0
                ymax = 1.0

            ax.set_ylim(ymin, ymax)
            '''

            if ncols >= 4:
                for l in ax.get_xticklabels():
                    l.set_fontsize(8)
            else:
                for l in ax.get_xticklabels():
                    l.set_fontsize(10)
            for l in ax.get_yticklabels():
                l.set_fontsize(10)

            if time_unit == 'h' and x.max() - x.min() <= 5.0:
                majorLocator = MultipleLocator(1.0)
                minorLocator = MultipleLocator(0.5)

                ax.xaxis.set_major_locator(majorLocator)
                ax.xaxis.set_minor_locator(minorLocator)

            if varScale == 'log' or varScale == 'log-log':
                ax.set_xscale('log')
                ax.set_yscale('log')
                positiveX = x > 0.0
                x = np.extract(positiveX, x)
                y = np.extract(positiveX, y)
                positiveY = y > 0.0
                x = np.extract(positiveY, x)
                y = np.extract(positiveY, y)
                if y.size > 0:
                    if y.min() > 0.0 and y.max() > y.min():
                        ymax = y.max()
                        dy = ymax * .1
                        ymax += dy
                        ymin = y.min()
                        ymin -= dy
                        if ymin < 0.0 or ymin > ymax / 1000.0:
                            ymin = ymax / 1000.0
                        ax.set_ylim(ymin, ymax)
                    else:
                        ax.set_ylim(1.0, 10.0)
                else:
                    ax.set_ylim(1.0, 10.0)

            if varScale == 'log-linear':
                ax.set_xscale('log')
                positiveX = x > 0.0 # True if > 0.0
                x = np.extract(positiveX, x)
                y = np.extract(positiveX, y)

            if varScale == 'linear-log':
                ax.set_yscale('log')
                positiveY = y > 0.0 # True if > 0.0
                x = np.extract(positiveY, x)
                y = np.extract(positiveY, y)
                #assert x.size == y.size, 'size error; stop.'
                if y.size > 0:
                    if y.min() > 0.0 and y.max() > y.min():
                        ymax = y.max()
                        dy = ymax * .1
                        ymax += dy
                        ymin = y.min()
                        ymin -= dy
                        if ymin < 0.0 or ymin > ymax / 1000.0:
                            ymin = ymax / 1000.0
                        ax.set_ylim(y.min(), ymax)
                    else:
                        ax.set_ylim(1.0, 10.0)
                else:
                    ax.set_ylim(1.0, 10.0)

            # ...................
            # make the plot here

            yformatter = ScalarFormatter(useMathText=True,useOffset=True)
            yformatter.set_powerlimits((15, 5))
            ax.yaxis.set_major_formatter(yformatter)

            if varLegend:

                ax.plot(x, y, 's-', color='black', linewidth=0.5, markersize=2,
                        markeredgecolor='black', label=varLegend)
            else:

                ax.plot(x, y, 's-', color='black', linewidth=0.5, markersize=2,
                        markeredgecolor='black')

            # ...................

            if species.info:
                ax.set_title(species.info,fontsize=8)
            if varLegend:
                ax.legend(loc='best', prop={'size': 7})
            ax.grid()

        # end of: for i_var in range(num_var):

        fig_name = name+'-'+self.name+'-phase-plot-' + str(i_dash).zfill(2)
        fig.savefig(fig_name+'.png', dpi=dpi, fomat='png')
        plt.close(fig_num)

        #pickle.dump( fig, open(fig_name+'.pickle','wb') )

        return

if __name__ == '__main__':
    tbp_org = Species( name='TBP', formula_name='(C4H9O)_3PO(o)',
              phase_name='organic', atoms=['12*C','27*H','4*O','P'] )
    quant = Quantity( name='volume' )
    phase = PhaseNew(name='solvent',species=[tbp_org],quantities=[quant])
    print(phase)



#!/usr/bin/env python
from cortix import Cortix
from cortix import Module
from cortix import Network
from cortix import Port
from body import Body

from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

import os
import numpy as np

def main():
    um = False
    sim_time = 365 * 24 * 3600 #157788000.0 * 10
    time_step = 24 * 3600 #25000.0

    universe_rad = 2.5e11

    cortix = Cortix(use_mpi=um)
    cortix.network = Network()

    earth_mass = 5.9740e+24
    earth_pos = np.array([1.4960e+11, 0.0, 0.0])
    earth_vel = np.array([(0.0, 2.9800e+04, 0.0)])
    earth = Body(earth_mass, earth_pos, earth_vel)
    earth.name = "earth"
    cortix.network.module(earth)

    mars_mass = 6.4190e+23
    mars_pos = np.array([2.2790e+11, 0.0, 0.0])
    mars_vel = np.array([0.0, 2.4100e+04, 0.0])
    mars = Body(mars_mass, mars_pos, mars_vel)
    mars.name = "mars"
    cortix.network.module(mars)

    mercury_mass = 3.3020e+23
    mercury_pos = np.array([5.7900e+10, 0.0, 0.0])
    mercury_vel = np.array([0.0, 4.7900e+04, 0.0])
    mercury = Body(mercury_mass, mercury_pos, mercury_vel)
    mercury.name = "mercury"
    cortix.network.module(mercury)

    venus_mass = 4.8690e+24
    venus_pos = np.array([1.0820e+11, 0.0, 0.0])
    venus_vel = np.array([0.0, 3.5000e+04 , 0.0])
    venus = Body(venus_mass, venus_pos, venus_vel)
    venus.name = "venus"
    cortix.network.module(venus)

    sun_mass = 1.9890e+30
    sun_pos = np.array([0.0, 0.0, 0.0])
    sun_vel = np.array([0.0, 0.0, 0.0])
    sun = Body(sun_mass, sun_pos, sun_vel)
    sun.name = "sun"
    cortix.network.module(sun)

    for x, i in enumerate(cortix.network.modules):
        i.save = True
        i.dt = time_step
        i.time = sim_time
        for y, j in enumerate(cortix.network.modules):
            pi = Port("body_{}".format(y), um)

            if x != y and pi not in i.ports:
                i.ports.append(pi)

                pj = Port("body_{}".format(x), um)
                j.ports.append(pj)

                pj.connect(pi)
                cortix.network.gv_edges.append((str(x), str(y)))

    cortix.run()
    cortix.network.draw()

    return [(b.name, b.trajectory) for b in cortix.network.modules]

def plot_trajectories(traj=None):
    au = (149.6e6 * 1000)
    scale = 250 / au

    if traj is None:
        traj = []
        for file in os.listdir("."):
            if file.endswith(".csv"):
                with open(file, "r") as f:
                    lines = [l.strip().split(",") for l in f.readlines()]
                    name = lines[0]
                    lines = line[1:]
                    t = [(float(i[0]), float(i[1]), float(i[2])) for i in lines]
                    traj.append((name, t))


    fig = plt.figure(1)
    ax = fig.gca(projection='3d')
    patches = []
    plt.draw()
    plt.pause(.1)
    plt.legend(handles=[Patch(label=name) for (name, t) in traj])

    for (name, t) in traj:
        patches.append(name)
        x = [i[0] * scale for i in t]
        y = [i[1] * scale for i in t]
        z = [i[2] * scale for i in t]
        ax.plot(x, y, z)

    plt.savefig("planets.png")


if __name__ == "__main__":
    trajectories = main()
    import matplotlib
    matplotlib.use("GTK3Agg")
    plot_trajectories(trajectories)
    plt.show()



from cortix.src.cortix_main import Cortix
from cortix.src.network import Network

from cortix.tests.dummy_module import DummyModule
from cortix.tests.dummy_module import DummyModule2

def mpi_send_recv():

    c = Cortix()

    c.network = Network()

    m1 = DummyModule()
    c.network.add_module(m1)
    p1 = m1.get_port('test-port1')

    m2 = DummyModule2()
    c.network.add_module(m2)
    p2 = m2.get_port('test-port2')

    c.network.connect([m1,p1],[m2,p2],'bidirectional')

    c.run()

if __name__ == "__main__":
    mpi_send_recv()




'''Parameters'''

params = dict()

#Data pertaining to one-group energy neutron balance
params['gen_time']     = 1.0e-4  # s
params['beta']         = 6.5e-3  # 
#params['diff_coeff']   = 0.84 # cm
#params['core radius']  = 1.8e+2 # cm
#params['core_height']  = 2.8e+2 # cm
#params['k_infty']      = 1.118
#params['Sigma_a']      = 2.74e-4 # 1/cm

#Delayed neutron emission
params['species_decay']     = [0.0124, 0.0305, 0.111, 0.301, 1.14, 3.01] # 1/sec
params['species_rel_yield'] = [0.033, 0.219, 0.196, 0.395, 0.115, 0.042]

params['alpha_n'] = -5e-4 # control rod reactivity worth
params['alpha_tn_fake'] = -1e-4/20# -1.0e-6

params['n_dens_ss_operation'] = 1e15 * 1e4 / 2200  # neutrons/m^2

#Data pertaining to two-temperature heat balances
params['fis_energy']           = 180 * 1.602e-13 # J/fission 
#params['enrich']               = 3./100.
#params['fuel_mat_mass_dens']   = 10.5 # g/cc
#params['moderator_fuel_ratio'] = 387 # atomic number concentration ratio
params['sigma_f_o']            = 586.2 * 100 * 1e-30 # m2
params['temp_o']               = 20 + 273.15 # K
params['temp_c_ss_operation']  = 550 + 273.15 # K desired ss operation temp of coolant
params['temp_f_safe_max']      = 1100 + 273.15 # K
params['thermal_neutron_velo'] = 2200 # m/s

params['fis_nuclide_num_dens_fake'] = 1e17/40 * 1.0e+6 # (fissile nuclei)/m3

#params['q_c'] = 0  # volumetric flow rate

params['fuel_dens']   = 2500 # kg/m3
params['cp_fuel']     = 720 # J/(kg K)
params['fuel_volume'] = 1.5 # m3

params['coolant_dens']   = 0.1786 #  kg/m3
params['cp_coolant']     = 20.78 / 4e-3 # J/(mol K) - > J/(kg K)
params['coolant_volume'] = 0.8 # m3

params['ht_coeff'] = 800 # W/K

#params['fis_prod_beta_energy_rate']  = 1.26 * 1.602e-13 # J/(fission sec) 1.26 t^-1.2 (t in seconds)
#params['fis_prod_gamma_energy_rate'] = 1.40 * 1.602e-13 # J/(fission sec) 1.40 t^-1.2 (t in seconds)

params['strict'] = True # apply strict testing to some quantities

# Operations
params['hx_malfunction'] = False
params['hx_malfunction_start_time']  = 0
params['hx_malfunction_down_time']   = 0
params['hx_malfunction_temperature'] = 0

params['hx_breakdown'] = False
params['hx_breakdown_start_time'] = 0
params['hx_breakdown_down_time']  = 0

params['shutdown']      = False
params['shutdown_time'] = 0.0 # s
params['rho_shutdown']  = 0.0 # s
params['hx_relax_time'] = 0 # s

'''Setup function for delayed neutron species concentrations at steady state'''

def setup_initial_conditions(params):

    # setup the steady state for the delayed-neutron precursors

    n_species = len(params['species_decay'])

    assert len(params['species_rel_yield']) == n_species

    import numpy as np
    c_vec_0 = np.zeros(n_species,dtype=np.float64) # initialize conentration vector

    species_decay = params['species_decay'] # retrieve list of decay constants
    lambda_vec    = np.array(species_decay) # create a numpy vector

    species_rel_yield = params['species_rel_yield']
    beta_vec = np.array(species_rel_yield) * beta  # create the beta_i's vector

    gen_time = params['gen_time'] # retrieve neutron generation time

    n_ss = params['n_ss']
    c_vec_ss = beta_vec/lambda_vec/gen_time * n_ss # compute the steady state precursors number density

    params['c_vec_ss'] = c_vec_ss

    # setup initial condition for variables
    params['n_0']     = n_ss
    params['c_vec_0'] = c_vec_ss
    params['rho_0']   = params['reactivity']

    params['temp_f_0'] = params['temp_0'] + 10.0 # helps startup integration
    params['temp_c_0'] = params['temp_0']

    return

'''Reactivity coefficient function'''

def alpha_tn_func(temp, params):
    '''
    Place holder for implementation
    '''
    alpha_tn = params['alpha_tn_fake']

    return alpha_tn

'''Reactivity function'''

def rho_func( time, n_dens, temp, params ):
    '''
    Reactivity function.

    Parameters
    ----------
    t: float, required
        Time.
    temp_f: float, required
        Temperature at time t.
    params: dict, required
        Dictionary of quantities. It must have a `'rho_0'` key/value pair.

    Returns
    -------
    rho_t: float
        Value of reactivity.

    Examples
    --------
    '''

    if params['shutdown'] == False or (params['shutdown'] == True and time < params['shutdown_time']):
        rho_0 = params['rho_0']
        n_dens_ref = params['n_dens_ref']
        temp_ref = params['temp_c_ss_operation']
        alpha_n  = params['alpha_n']
        alpha_tn = alpha_tn_func( temp, params )
    elif params['shutdown'] == True and time >= params['shutdown_time']:
        rho_0 = params['rho_shutdown']
        n_dens_ref = 0.0
        temp_ref = params['temp_o']
        alpha_n  = 0
        alpha_tn = 0
    else:
        assert False

    rho_t = rho_0 + alpha_n * (n_dens - n_dens_ref) + alpha_tn * (temp - temp_ref)

    beta = params['beta']

    if params['strict'] == True:
        assert rho_t/beta < 1,'rho/beta = %r at time = %r; rho_0 = %r; rho_n = %r; rho_tn = %r'%                ( rho_t/beta, time, rho_0/beta, alpha_n*(n_dens - n_dens_ref)/beta,                 alpha_tn*(temp - temp_ref)/beta )

    return rho_t

'''Source function'''

def q_source( t, params ):
    '''
    Neutron source delta function.

    Parameters
    ----------
    t: float, required
        Time.
    params: dict, required
        Dictionary of quantities. It must have a `'q_0'` key/value pair.

    Returns
    -------
    q: float
        Value of source.

    Examples
    --------
    '''

    q = 0.0
    q_0 = params['q_0']

    if t <= 1e-5: # small time value
        q = q_0
    else:
        q = 0.0

    return q

'''Effective microscopic fission cross section'''

def sigma_fis_func( temp, params ):
    '''
    Place holder for implementation
    '''

    sigma_f_fake = params['sigma_f_o']

    return sigma_f_fake

'''Nuclear heating power density function'''

def nuclear_pwr_dens_func( time, temp, n_dens, params ):
    '''
    Place holder for implementation
    '''

    rxn_heat = params['fis_energy'] # get fission reaction energy J per reaction

    sigma_f = sigma_fis_func( temp, params ) # m2

    fis_nuclide_num_dens = params['fis_nuclide_num_dens_fake'] #  #/m3

    Sigma_fis = sigma_f * fis_nuclide_num_dens # macroscopic cross section

    v_o = params['thermal_neutron_velo'] # m/s

    neutron_flux = n_dens * params['n_dens_ss_operation'] * v_o

    # reaction rate density
    rxn_rate_dens = Sigma_fis * neutron_flux

    # nuclear power source
    q3prime = - rxn_heat * rxn_rate_dens # exothermic reaction W/m3

    if params['strict'] == True:
        assert q3prime <= 0.0,"time = %r, q''' = %r, n_dens = %r"%(time,q3prime,n_dens)

    return q3prime

'''Cooling heat rate'''

def heat_sink_rate( time, temp_f, temp_c, params):

    ht_coeff = params['ht_coeff']

    q_f = - ht_coeff * (temp_f - temp_c)

    if params['strict'] == True:
        assert q_f <= 0.0,'q_f = %r at time = %r; temp_f = %r, temp_c = %r'%(q_f,time,temp_f,temp_c)

    return q_f

'''Coolant Inlet Temperature'''

def coolant_inlet_temp_func( time, temp_f, temp_c, params ):

    import math

    if params['shutdown'] == False or (params['shutdown'] == True and time < params['shutdown_time']):
        if temp_c < params['temp_c_ss_operation']:
            temp_in = temp_c
        else:
            temp_in = params['temp_c_ss_operation']
    elif params['shutdown'] == True and time >= params['shutdown_time']:
        temp_in = params['temp_o'] +                   ( params['temp_c_ss_operation'] - params['temp_o'] ) * math.exp(-(time-params['shutdown_time'])/params['hx_relax_time'])
    else:
        assert False

    # protect the fuel
    if params['strict'] == True:
        assert temp_f <= params['temp_f_safe_max'],'Above max. fuel temp. = %r [C], at time = %r [h]'%            (temp_f-273.15,time/3600)

    if params['hx_malfunction'] == True:
        start_time = params['hx_malfunction_start_time']
        end_time   = start_time + params['hx_malfunction_down_time']
        if time >= start_time and time <= end_time:
            temp_in = params['hx_malfunction_temperature'] # external heat exchanger malfunction

    if params['hx_breakdown'] == True:
        start_time = params['hx_breakdown_start_time']
        end_time   = start_time + params['hx_breakdown_down_time']
        if time >= start_time and time <= end_time:
            temp_in = temp_c         # external heat exchanger breakdown

    return temp_in

'''ODE function'''

def f_vec( time, u_vec, params ):  
    
    import numpy as np
    #assert np.all(u_vec >= 0.0),'time = %r; u_vec = %r'%(time,u_vec)

    n_dens = u_vec[0] # get neutron dens

    c_vec = u_vec[1:-2] # get delayed neutron emitter concentration

    temp_f = u_vec[-2] # get temperature of fuel

    temp_c = u_vec[-1] # get temperature of coolant

    # initialize f_vec to zero
    species_decay = params['species_decay']
    lambda_vec = np.array(species_decay)
    n_species  = len(lambda_vec)

    f_tmp = np.zeros(1+n_species+2,dtype=np.float64) # vector for f_vec return

    #----------------
    # neutron balance
    #----------------
    rho_t    = rho_func(time, n_dens, (temp_f+temp_c)/2.0, params)

    beta     = params['beta']
    gen_time = params['gen_time']

    species_rel_yield = params['species_rel_yield']
    beta_vec = np.array(species_rel_yield) * beta

    assert len(lambda_vec)==len(beta_vec)

    q_source_t = q_source(time, params)

    f_tmp[0] = (rho_t - beta)/gen_time * n_dens + lambda_vec @ c_vec + q_source_t

    #-----------------------------------
    # n species balances (implicit loop)
    #-----------------------------------
    f_tmp[1:-2] = beta_vec/gen_time * n_dens - lambda_vec * c_vec

    #--------------------
    # fuel energy balance
    #--------------------
    rho_f    = params['fuel_dens']
    cp_f     = params['cp_fuel']
    vol_fuel = params['fuel_volume']

    pwr_dens = nuclear_pwr_dens_func( time, (temp_f+temp_c)/2, n_dens, params )

    heat_sink = heat_sink_rate( time, temp_f, temp_c, params )

    f_tmp[-2] = - 1/rho_f/cp_f * ( pwr_dens - heat_sink/vol_fuel )

    #-----------------------
    # coolant energy balance
    #-----------------------
    rho_c    = params['coolant_dens']
    cp_c     = params['cp_coolant']
    vol_cool = params['coolant_volume']

    temp_in = coolant_inlet_temp_func(time, temp_f, temp_c, params)

    tau = params['tau_fake']

    heat_source = - heat_sink

    f_tmp[-1] = - 1/tau * (temp_c - temp_in) + 1./rho_c/cp_c/vol_cool * heat_source

    return f_tmp


'''Create the point-reactor run function'''

def run_point_reactor( f_vec, params ):

    from scipy.integrate import odeint # Load ODE solver package

    import numpy as np
    time_final = params['time_final']
    n_time_stamps = params['n_time_stamps']
    time_stamps = np.linspace(0.0, time_final, num=n_time_stamps) # create the time stamps for solution values
    params['time_stamps'] = time_stamps

    max_n_steps_per_time_step = 200 # max number of nonlinear algebraic solver iterations per time step

    n_0     = params['n_0']
    c_vec_0 = params['c_vec_0']

    temp_f_0 = params['temp_f_0']
    temp_c_0 = params['temp_c_0']

    # m-equation point reactor model
    n_species = len(c_vec_0)
    u_vec_0   = np.zeros(1+n_species+2,dtype=np.float64)

    u_vec_0[0]    = n_0
    u_vec_0[1:-2] = c_vec_0
    u_vec_0[-2]   = temp_f_0
    u_vec_0[-1]   = temp_c_0

    #rtol = np.ones(u_vec_0.size)*1e-1
    #atol = np.ones(u_vec_0.size)
    #for i in range(u_vec_0.size-2):
        #print(i)
        #atol[i] = max(u_vec_0[i]*1e-1,1e-4)
        #print(atol)

    (u_vec_history, info_dict) = odeint( f_vec, u_vec_0, time_stamps,
                                         args=( params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=max_n_steps_per_time_step,
                                         #rtol=rtol, atol=atol, mxstep=max_n_steps_per_time_step,
                                         #hmax=2,
                                         full_output=True, tfirst=True )

    assert info_dict['message']=='Integration successful.',                     'Fatal: scipy.integrate.odeint failed %r'%info_dict['message']


    return u_vec_history

'''Plotting function definition'''

def plot_neutron_results( u_vec_history, normalize=False, semi_log=False, markers=False, precursors=True ):

    time_stamps = params['time_stamps']
    n_dens_ss = params['n_dens_ss_operation']
    c_vec_normalize = u_vec_history[-1,1:-2]

    import matplotlib.pyplot as plt

    fig, ax1 = plt.subplots(1, figsize=(14, 6))

    if precursors == True:

        ax2 = ax1.twinx() # duplicate x axes to plot n and c_i's in different y axes

        color_ids = np.linspace(0,1,u_vec_history[:,1:-2].shape[1])

        for (j,color_id) in zip( range(u_vec_history[:,1:-2].shape[1]), color_ids ):
            color=plt.cm.nipy_spectral(color_id)

            if normalize == True:
                ax2.plot( time_stamps/3600,u_vec_history[:,j+1]/c_vec_normalize[j],'-.',color=color,label=r'$c_%i$'%(j+1) )
                ax2.set_ylabel(r'$c_i/c_{i_0}$',fontsize=16,color='black')
            else:
                ax2.plot( time_stamps/3600,u_vec_history[:,j+1],'-.',color=color,label=r'$c_%i$'%(j+1) )
                ax2.set_ylabel(r'$c_i$',fontsize=16,color='black')

        ax2.tick_params(axis='y', labelcolor='black', labelsize=14)
        ax2.legend(loc='lower right',fontsize=12)
        if semi_log == True:
            ax2.set_yscale('log') # uncomment to plot y in log scale
        #ax2.grid(True)

    if markers == True:
        if normalize == True:
            ax1.plot( time_stamps/3600,u_vec_history[:,0]/n_dens_ss,'-',marker='+',color='red',label=r'$n/n_0$' )
            ax1.set_ylabel(r'$n$',fontsize=16,color='black')
        else:
            ax1.plot( time_stamps/3600,u_vec_history[:,0],'-',marker='+',color='red',label=r'$n$' )
            ax1.set_ylabel(r'$n$',fontsize=16,color='black')
    else:
        if normalize == True:
            ax1.plot(time_stamps/3600,u_vec_history[:,0]/n_dens_ss,'-',color='red',label=r'$n/n_0$' )
            ax1.set_ylabel(r'$n/n_{ss}$',fontsize=16,color='black')
        else:
            ax1.plot(time_stamps/3600,u_vec_history[:,0],'-',color='red',label=r'$n$' )
            ax1.set_ylabel(r'$n$',fontsize=16,color='black')

    ax1.set_xlabel(r'Time [h] ($\tau=$%4.3f s)'%tau,fontsize=16)

    ax1.tick_params(axis='y', labelcolor='black', labelsize=14)
    ax1.tick_params(axis='x', labelsize=14)
    ax1.legend(loc='lower left',fontsize=12)
    if semi_log == True:
        ax1.set_yscale('log') # uncomment to plot y in log scale
    ax1.grid(True)

    plt.title(r'Point-Reactor Model: $\rho/\beta=$'
              +str(params['reactivity']/params['beta'])
              +r'; $q_0=$'+str(round(params['q_0'],2)),
              fontsize=18)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.show()

    print('')

'''Utility function to peak at results table'''

def peek(time,data, head=5, tail=5):  

    import pandas as pd
    
    pd.options.display.float_format = '{:.2e}'.format
    
    layout = {'time':time[:head]}
    
    layout['n'] = data[:head,0]
    
    for j in range(1,data[:,1:-2].shape[1]+1):
        layout['c_%i'%j] = data[:head,j]

    layout['temp_f [K]'] = data[:head,-2]
    layout['temp_c [K]'] = data[:head,-1]

    results = pd.DataFrame(layout)
    print(round(results,2))
    print('')

    layout = {'time':time[-tail:]}

    layout['n'] = data[-tail:,j]

    for j in range(1,data[:,1:-2].shape[1]+1):
        layout['c_%i'%j] = data[-tail:,j]

    layout['temp_f [K]'] = data[-tail:,-2]
    layout['temp_c [K]'] = data[-tail:,-1]

    results = pd.DataFrame(layout)
    print(round(results,2))
    print('')

'''Post-Processing Quantities'''

def get_quantities(u_vec_history,params):

    import pandas as pd
    data = dict()
    tmp1 = list()
    tmp2 = list()
    tmp3 = list()
    tmp4 = list()

    for (time,n_dens,tf,tc) in zip(params['time_stamps'],u_vec_history[:,0],u_vec_history[:,-2],u_vec_history[:,-1]):

        rho = rho_func(time,n_dens,(tf+tc)/2,params)
        tmp1.append( rho )

        temp_in = coolant_inlet_temp_func( time, tf, tc, params )
        tmp2.append(temp_in)

        temp_ave = (tf+tc)/2.0
        q3prime = nuclear_pwr_dens_func( time, temp_ave, n_dens, params )
        tmp4.append(q3prime/1000)

        pwr = params['coolant_volume']/params['tau_fake'] * params['coolant_dens'] * params['cp_coolant'] * (tc - temp_in)
        tmp3.append(pwr/1000)

    data['rho'] = tmp1
    data['T_in [C]'] = np.array(tmp2)-273.15
    data['pwr [kW]'] = tmp3
    data["q''' [kW/m3]"] = tmp4

    df = pd.DataFrame(data)

    return df

'''Plot Tf and Tc temperatures in the reactor'''

def plot_temperature_results( u_vec_history,params, semi_log=False ):

    quant = get_quantities(u_vec_history,params)

    time_stamps = params['time_stamps']

    import matplotlib.pyplot as plt
    fig, ax1 = plt.subplots(1, figsize=(16, 6))
    #ax1.plot(time_stamps/3600,u_vec_history[:,-1]-273.15,'r-',time_stamps/3600,u_vec_history[:,-2]-273.15,'b-',label='$T_f=$')

    ax1.plot(time_stamps/3600,u_vec_history[:,-2]-273.15,'r-.',label='$T_f$')
    ax1.plot(time_stamps/3600,u_vec_history[:,-1]-273.15,'r-',label='$T_c$')
    ax1.plot(time_stamps/3600,quant['T_in [C]'],'b-',label=r'$T_{in}$' )

    ax1.set_xlabel(r'Time [h] ($\tau=$%4.3f s)'%tau,fontsize=16)
    ax1.set_ylabel(r'$T$ [C]',fontsize=16,color='black')
    ax1.tick_params(axis='y', labelcolor='black', labelsize=14)
    ax1.tick_params(axis='x', labelsize=14)
    ax1.legend(loc='upper left',fontsize=12)
    ax1.grid(True)
    if semi_log == True:
        ax1.set_yscale('log')

    plt.title('Single-Point Reactor HTTR Temperatures',fontsize=20)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.show()
    print('')

'''Reactor Power and Delta T (input/output) at Steady State'''

def get_ss_system_pwr(u_vec_history,params):

    quant = get_quantities(u_vec_history,params)

    temp_in = quant['T_in [C]'].iat[-1]
    delta_T = (u_vec_history[-1,-1]-273.15 - temp_in)

    pwr = quant['pwr [kW]'].iat[-1]

    return (pwr,delta_T)

'''Plot Reactivity'''

def plot_reactivity(u_vec_history,params):

    quant = get_quantities(u_vec_history,params)

    time_stamps = params['time_stamps']
    import matplotlib.pyplot as plt
    fig, ax1 = plt.subplots(1, figsize=(16, 6))
    ax1.plot(time_stamps/3600,quant['rho']/params['beta'],'b-',label=r'$\rho$' )

    ax1.set_xlabel(r'Time [h] ($\tau=$%4.3f s)'%tau,fontsize=16)
    ax1.set_ylabel(r'$\rho(t)/\beta$',fontsize=16,color='black')
    ax1.tick_params(axis='y', labelcolor='black', labelsize=14)
    ax1.tick_params(axis='x', labelsize=14)
    ax1.legend(loc='upper left',fontsize=12)
    ax1.grid(True)

    plt.title('Single-Point Reactor HTTR Reactivity',fontsize=20)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.show()
    print('')

    return

'''Plot Heating Power'''

def plot_heating_power(u_vec_history,params):

    quant = get_quantities(u_vec_history,params)

    time_stamps = params['time_stamps']
    import matplotlib.pyplot as plt
    fig, ax1 = plt.subplots(1, figsize=(16, 6))

    ax1.plot(time_stamps/3600,quant['pwr [kW]'],'-',color='black',label='$Pwr$' )

    ax1.set_xlabel(r'Time [h] ($\tau=$%4.3f s)'%tau,fontsize=16)
    ax1.set_ylabel(r'Pwr$_{th}$ [kW]',fontsize=16,color='black')
    ax1.tick_params(axis='y', labelcolor='black', labelsize=14)
    ax1.tick_params(axis='x', labelsize=14)
    ax1.legend(loc='upper left',fontsize=12)
    ax1.xaxis.grid(True,linestyle='-',which='major',color='lightgrey',alpha=0.9)
    #ax1.grid(True)

    ax2 = ax1.twinx()

    ax2.plot(time_stamps/3600,-quant["q''' [kW/m3]"]*params['fuel_volume'],'r-.',label=r"$q^{'''}$" )

    ax2.set_ylabel(r"$-q^{'''}\,V_f$ [kW]",fontsize=18,color='red')
    ax2.tick_params(axis='y', labelcolor='red', labelsize=14)
    ax2.legend(loc='lower right',fontsize=12)
    ax2.set_ylim(0,max(ax2.get_ylim()))
    #ax2.grid(True)

    ax1.set_ylim(0,max(ax2.get_ylim()))

    plt.title('Single-Point Reactor HTTR Removed Heat Rate History',fontsize=20)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.show()
    print('')
    return

'''Decay Heating Power Density'''

def decay_heat_pwr_dens(u_vec_history,params):

    if params['shutdown'] == False:
        print("params['shutdown'] is False; bailing out.")
        return None

    import pandas as pd
    import scipy.integrate as integrate

    time_shutdown = params['shutdown_time']
    time_final    = params['time_final']
    time_cooling  = time_final - time_shutdown

    decay_time = list()
    decay_pwr  = list()

    n_pts = 150
    for x in range(n_pts+1):

        if x <= 2:  # avoid the case with no cooling period because the time singularity
            continue

        time_cooling = x * (time_final-time_shutdown)/n_pts

        tmp1 = list()
        tmp2 = list()

        for (time,n_dens,tf,tc) in zip(params['time_stamps'],u_vec_history[:,0],u_vec_history[:,-2],u_vec_history[:,-1]):

            if time > time_shutdown:
                break

            temp = (tf+tc)/2
            sigma_f = sigma_fis_func( temp, params ) # m2

            fis_nuclide_num_dens = params['fis_nuclide_num_dens_fake'] #  #/m3

            Sigma_fis = sigma_f * fis_nuclide_num_dens # macroscopic cross section

            v_o = params['thermal_neutron_velo'] # m/s

            neutron_flux = n_dens * params['n_dens_ss_operation'] * v_o

            assert time_shutdown - time + time_cooling > 0.0

            integrand = 4.2618e-13 * Sigma_fis * neutron_flux * (time_shutdown - time + time_cooling)**(-1.2)

            tmp1.append(time)
            tmp2.append(integrand)

        integral = integrate.trapz(tmp1,tmp2)
        #print('time [h]             = ',(time_cooling+time_shutdown)/3600)
        #print('decay heat power [W] = ',round(integral*params['fuel_volume'],3))

        decay_time.append(time_cooling+time_shutdown)
        decay_pwr.append(integral*params['fuel_volume'])

    data = dict()
    data['time [s]']  = decay_time
    data['decay [W]'] = decay_pwr

    df = pd.DataFrame(data)

    return df

'''Plot Heating Power'''

def plot_decay_heating_power(u_vec_history,decay_heat,params):

    if params['shutdown'] == False:
        print("params['shutdown'] is False; bailing out.")
        return None

    quant = get_quantities(u_vec_history,params)

    time_stamps = params['time_stamps']

    import matplotlib.pyplot as plt
    fig, ax1 = plt.subplots(1, figsize=(16, 6))

    ax1.plot(decay_heat['time [s]']/3600,decay_heat['decay [W]']/1000,'-',color='red',label='Decay Pwr' )

    ax1.set_xlabel(r'Time [h] ($\tau=$%4.3f s)'%tau,fontsize=16)
    ax1.set_ylabel(r'Decay Heating Power [kW]',fontsize=16,color='red')
    ax1.tick_params(axis='y', labelcolor='red', labelsize=14)
    ax1.tick_params(axis='x', labelsize=14)
    ax1.legend(loc='upper left',fontsize=12)
    ax1.xaxis.grid(True,linestyle='-',which='major',color='lightgrey',alpha=0.9)

    #ax1.grid(True)

    ax2 = ax1.twinx()
    ax2.plot(time_stamps/3600,-quant["q''' [kW/m3]"]*params['fuel_volume'],'k-.',label=r"$-q^{'''}\,V_f$" )
    ax2.set_ylabel(r"$-q^{'''}\,V_f$ [kW]",fontsize=18,color='black')
    ax2.tick_params(axis='y', labelcolor='black', labelsize=14)
    ax2.legend(loc='lower right',fontsize=12)
    ax2.set_ylim(0,max(ax2.get_ylim()))
    #ax2.grid(True)

    ax1.set_ylim(0,max(ax2.get_ylim()))

    plt.title('Single-Point Reactor HTTR Decay Heat History',fontsize=20)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    plt.show()
    print('')
    return

'''Setup up initial conditions'''

import numpy as np

gen_time = params['gen_time'] # retrieve neutron generation time
params['q_0'] = 1/gen_time # pulse at t = 0

params['n_ss']       = 0.0 # neutronless steady state before start up
params['n_dens_ref'] = 1.0

rho_0_over_beta = 0.5 # $

beta = params['beta'] # retrieve the delayed neutron fraction
params['reactivity'] = rho_0_over_beta * beta # "rho/beta = 10 cents"

params['temp_0'] = params['temp_o']

params['tau_fake'] = .025 # s residence time

# setup remaining initial conditions
setup_initial_conditions( params )

'''Evolve the point-reactor'''

time_final    = 3600 * 25 # s
n_time_stamps = 500 # number of solution values in time

params['time_final']    = time_final
params['n_time_stamps'] = n_time_stamps

# Run the reactor and compute the history of the state variables; tabular form, one row per time stamp
u_vec_history = run_point_reactor( f_vec, params )

peek(params['time_stamps'],u_vec_history)

'''Compute Coolant Flow Rate: DELETE ME'''

tau = params['tau_fake']
flow_rate_c = params['coolant_volume'] * 1000 / (tau*60)
print('coolant vol flow rate [L/min] = ',flow_rate_c)

'''Plot neutron and delayed neutron emitter concentration in the reactor'''

plot_neutron_results(u_vec_history, semi_log=False, normalize=False, precursors=True)

'''Plot Reactivity'''

plot_reactivity(u_vec_history,params)

'''Plot Temperatures'''

plot_temperature_results(u_vec_history,params,semi_log=False)

'''Plot Heating Powers'''

plot_heating_power(u_vec_history,params)

'''Heating Power and Delta T'''

(pwr_kw,delta_T_C) = get_ss_system_pwr(u_vec_history,params)

print('Power = %.2e [kW]'%pwr_kw)
print('Delta T = %.2e [C]'%delta_T_C)

'''Add conditions for HX malfunction'''

params['strict'] = False
params['hx_malfunction'] = True

params['hx_malfunction_start_time']  = 8 * 3600
params['hx_malfunction_down_time']   = 2.5 * 3600
params['hx_malfunction_temperature'] = 873 + 273.15

time_final    = 3600 * 25 # s, longer run to account for the disturbance transient
n_time_stamps = 500 # number of solution values in time

params['time_final']    = time_final
params['n_time_stamps'] = n_time_stamps

# Run the reactor and compute the history of the state variables; tabular form, one row per time stamp
u_vec_history = run_point_reactor( f_vec, params )

'''Plot neutron and delayed neutron emitter concentration in the reactor'''

plot_neutron_results(u_vec_history, semi_log=False, normalize=False, precursors=True)

'''Plot Reactivity'''

plot_reactivity(u_vec_history,params)

'''Plot Temperatures'''

plot_temperature_results(u_vec_history,params,semi_log=False)

'''Plot Heating Powers'''

plot_heating_power(u_vec_history,params)

'''Heating Power and Delta T'''

(pwr_kw,delta_T_C) = get_ss_system_pwr(u_vec_history,params)

print('Power = %.2e [kW]'%pwr_kw)
print('Delta T = %.2e [C]'%delta_T_C)

'''Revert to Default'''

params['hx_malfunction'] = False
params['strict'] = True

'''Add conditions for HX breakdown'''

params['strict'] = False
params['hx_breakdown'] = True

params['hx_breakdown_start_time'] = 11.8 * 3600
params['hx_breakdown_down_time']  = 4.0 * 3600

time_final    = 3600 * 25 # s, longer run to account for the disturbance transient
n_time_stamps = 500 # number of solution values in time

params['time_final']    = time_final
params['n_time_stamps'] = n_time_stamps

# Run the reactor and compute the history of the state variables; tabular form, one row per time stamp
u_vec_history = run_point_reactor( f_vec, params )

'''Plot neutron and delayed neutron emitter concentration in the reactor'''

plot_neutron_results(u_vec_history, semi_log=False, normalize=False, precursors=True)

'''Plot Reactivity'''

plot_reactivity(u_vec_history,params)

'''Plot Temperatures'''

plot_temperature_results(u_vec_history,params,semi_log=False)

'''Plot Heating Powers'''

plot_heating_power(u_vec_history,params)

'''Heating Power and Delta T'''

(pwr_kw,delta_T_C) = get_ss_system_pwr(u_vec_history,params)

print('Power = %.2e [kW]'%pwr_kw)
print('Delta T = %.2e [C]'%delta_T_C)

'''Revert to Default'''

params['hx_breakdown'] = False
params['strict']       = True


# ## Reactor Startup/Shutdown w/ Decay Heat: Reactivity Step<a id="shutdown"></a>
# 
# Apply a shutdown reactivity after steady state operation.

'''Setup up initial conditions'''

import numpy as np

gen_time = params['gen_time'] # retrieve neutron generation time
params['q_0'] = 1/gen_time # pulse at t = 0

params['n_ss'] = 0.0 # neutronless steady state before start up
params['n_dens_ref'] = 1.0

rho_0_over_beta = 0.5 # $

beta = params['beta'] # retrieve the delayed neutron fraction
params['reactivity'] = rho_0_over_beta * beta # "rho/beta = 10 cents"

params['temp_0'] = params['temp_o']

params['tau_fake'] = .025 # s residence time

params['shutdown']      = True
params['rho_shutdown']  = -.1* beta
params['shutdown_time'] = 3600 * 24

params['strict'] = False

params['hx_relax_time'] = 1.5 * 3600 # s

# setup remaining initial conditions
setup_initial_conditions(params)

'''Evolve the point-reactor'''

time_final    = 3600 * 28 # s, longer run to account for the disturbance transient
n_time_stamps = 500 # number of solution values in time

params['time_final']    = time_final
params['n_time_stamps'] = n_time_stamps

# Run the reactor and compute the history of the state variables; tabular form, one row per time stamp
u_vec_history = run_point_reactor( f_vec, params )

'''Compute Coolant Flow Rate: DELETE ME'''

tau = params['tau_fake']
flow_rate_c = params['coolant_volume'] * 1000 / (tau*60)
print('coolant vol flow rate [L/min] = ',flow_rate_c)

'''Plot neutron and delayed neutron emitter concentration in the reactor'''

plot_neutron_results(u_vec_history, semi_log=False, normalize=False, precursors=True)

'''Plot Reactivity'''

plot_reactivity(u_vec_history,params)

'''Plot Temperatures'''

plot_temperature_results(u_vec_history,params,semi_log=False)

'''Plot Heating Powers'''

plot_heating_power(u_vec_history,params)

df = decay_heat_pwr_dens(u_vec_history,params)

plot_decay_heating_power(u_vec_history,df,params)

params['hx_relax_time'] = 0
params['shutdown']      = False
params['rho_shutdown']  = 0
params['shutdown_time'] = 0

df['time [s]'] = df['time [s]']/3600

df.columns = ['time [h]','decay [kW]']

df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''
This example uses two modules instantiated many times. It be executed with MPI
(if `mpi4py` is available) or with the Python multiprocessing library. These choices
are made by variables listed below in the executable portion of this run file.

To run this case using MPI you should compute the number of
processes as follows:

    `nprocs = n_droplets + 1 vortex + 1 cortix`

then issue the MPI run command as follows (replace `nprocs` with a number):

     `mpiexec -n nprocs run_droplet.py`

To run this case with the Python multiprocessing library, just run this file at the
command line as

    `run_droplet.py`

'''

import scipy.constants as const


from cortix import Cortix
from cortix import Network
from cortix.examples.droplet_swirl.droplet import Droplet
from cortix.examples.droplet_swirl.vortex import Vortex


def main():
    '''Cortix run file for a `Droplet`-`Vortex` network.

    Attributes
    ----------
    n_droplets: int
        Number of droplets to use (one per process).
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    create_plots: bool
        Create various plots and save to files. (all data collected in the
        parent process; it may run out of memory).
    plot_vortex_profile: bool
        Whether to plot (to a file) the vortex function used.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    '''

    # Configuration Parameters
    n_droplets = 5
    end_time   = 3*const.minute
    time_step  = 0.2

    create_plots = True

    if n_droplets >= 2000:
        create_plots = False

    plot_vortex_profile = False # True may crash the X server.

    use_mpi = False # True for MPI; False for Python multiprocessing

    swirl = Cortix(use_mpi=use_mpi, splash=True)

    swirl.network = Network()

    # Vortex module (single).
    vortex = Vortex()
    swirl.network.module(vortex)
    vortex.show_time = (True,1*const.minute)
    vortex.end_time = end_time
    vortex.time_step = time_step
    if plot_vortex_profile:
        vortex.plot_velocity()

    for i in range(n_droplets):

        # Droplet modules (multiple).
        droplet = Droplet()
        swirl.network.module(droplet)
        droplet.end_time = end_time
        droplet.time_step = time_step
        droplet.bounce = False
        droplet.slip = False
        droplet.save = True

        # Network port connectivity (connect modules through their ports)
        swirl.network.connect( [droplet,'external-flow'],
                               [vortex,vortex.get_port('fluid-flow:{}'.format(i))],
                               'bidirectional' )

    swirl.network.draw()

    swirl.run()

    # Plot all droplet trajectories
    if create_plots:

        modules = swirl.network.modules

        if swirl.use_multiprocessing or swirl.rank == 0:

            # All droplets' trajectory

            from mpl_toolkits.mplot3d import Axes3D
            import matplotlib.pyplot as plt

            positions = list()
            for m in swirl.network.modules[1:]:
                positions.append(m.liquid_phase.get_quantity_history('position')[0].value)

            fig = plt.figure(1)
            ax = fig.add_subplot(111,projection='3d')
            ax.set_title('Droplet Trajectories')
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            for pos in positions:
                x = [i[0] for i in pos]
                y = [i[1] for i in pos]
                z = [i[2] for i in pos]
                ax.plot(x, y, z)
            fig.savefig('trajectories.png', dpi=300)

            # All droplets' speed
            fig = plt.figure(2)
            plt.xlabel('Time [min]')
            plt.ylabel('Speed [m/s]')
            plt.title('All Droplets')

            for m in modules[1:]:
                speed = m.liquid_phase.get_quantity_history('speed')[0].value
                plt.plot(list(speed.index/60), speed.tolist())

            plt.grid()
            fig.savefig('speeds.png', dpi=300)

            # All droplets' radial position
            fig = plt.figure(3)
            plt.xlabel('Time [min]')
            plt.ylabel('Radial Position [m]')
            plt.title('All Droplets')

            for m in modules[1:]:
                radial_pos = m.liquid_phase.get_quantity_history('radial-position')[0].value
                plt.plot(list(radial_pos.index/60)[1:], radial_pos.tolist()[1:])

            plt.grid()
            fig.savefig('radialpos.png', dpi=300)

    # This properly ends the program
    swirl.close()

if __name__ == '__main__':
    main()



#simple script to test the condenser
from params import get_params
from condenser import Condenser

input = get_params()
input['start-time'] = 0
input['end-time'] = 0

condenser = Condenser(input)
condenser.tester()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
from cortix import Module
from cortix import Port
from cortix import Network
from cortix import Cortix

from cortix.tests.dataplot import DataPlot

class PlotData(Module):
    def __init__(self):
        super().__init__()

    def run(self):
        i = 0
        while i < 10:
            data = (i, i**2)
            self.send(data, "plot-out")
            print("Sent {}!".format(data))
            i += 1
        self.send("DONE", "plot-out")
        print("Finished sending!")

if __name__ == "__main__":

    # Cortix built-in DataPlot module
    d = DataPlot()
    #d.set_title("Test Plot")
    #d.set_xlabel("Time")
    #d.set_ylabel("Position")

    d.title = 'Test Plot'
    d.xlabel = 'Time'
    d.ylabel = 'Position'

    # Custom class to send dummy data
    p = PlotData()

    p1 = Port("plot-in")
    p2 = Port("plot-out")
    p1.connect(p2)

    d.ports.append(p1)
    p.ports.append(p2)

    c = Cortix()
    c.network = Network()
    c.network.add_module(d)
    c.network.add_module(p)
    c.use_mpi = False
    c.run()



# startup params

import math
import scipy.constants as unit
import numpy as np

def get_params():
    params = dict()

    params['valve_opened'] = False
    params['valve_opening_time'] = 0

    params['operating-mode'] = 'startup'

    #initial SS values for reactor startup
    params['turbine-outflow-temp'] = 293.15
    params['turbine-chi'] = 0
    params['turbine-inlet-temp'] = 293.15
    params['condenser-runoff-temp'] = 287.15
    params['n-dens'] = 0
    params['fuel-temp'] = 293.15
    params['coolant-temp'] = 293.15
    params['delayed-neutron-cc'] = [0, 0, 0, 0, 0, 0]
    params['turbine-work'] = 0
    params['initial-flowrate'] = 0

    params['steam flowrate'] = 1820 #kg/s

    params['malfunction start'] = 999 * unit.hour
    params['malfunction end'] = 999 * unit.hour
    params['shutdown time'] = 999 * unit.hour

    params['temp_f_0'] = 300
    params['temp_c_0'] = params['coolant-temp']
    params['pressure_0'] = 1.013 # bar
    params['turbine-runoff-pressure'] = 1
    params['runoff-pressure'] = params['turbine-runoff-pressure']

    return(params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix import Phase
from cortix import Quantity

class Parole(Module):
    '''
    Parole Cortix module used to model criminal group population in a parole system.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `prison` and `community`.
    See instance attribute `port_names_expected`.

    '''

    def __init__(self, n_groups=1, pool_size=0.0):

        super().__init__()

        self.port_names_expected = ['prison','community']

        quantities      = list()
        self.ode_params = dict()

        self.initial_time = 0.0 * unit.day
        self.end_time     = 100 * unit.day
        self.time_step    = 0.5 * unit.day

        unit.percent = 1/100

        # Population groups
        self.n_groups = n_groups

        # Parole population groups
        feg_0 = np.random.random(self.n_groups) * pool_size
        feg = Quantity(name='feg', formal_name='parole-pop-grps',
                latex_name = '$n_e^{(g)}$',
                unit='# offenders', value=feg_0, info='Parole Population Groups')
        quantities.append(feg)

        # Model parameters: commitment coefficients

        # Parole to community
        a =  50*unit.percent/unit.year * np.ones(self.n_groups)
        b =  70*unit.percent/unit.year * np.ones(self.n_groups)
        ce0g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        ce0g = Quantity(name='ce0g', formal_name='commit-community-coeff-grps',
               unit='1/s', value=ce0g_0)
        self.ode_params['commit-to-community-coeff-grps'] = ce0g_0
        quantities.append(ce0g)

        # Parole to prison  
        a = 30*unit.percent/unit.year * np.ones(self.n_groups)
        b = 40*unit.percent/unit.year * np.ones(self.n_groups)
        cepg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        cepg = Quantity(name='cepg', formal_name='commit-prison-coeff-grps',
               unit='1/s', value=cepg_0)
        self.ode_params['commit-to-prison-coeff-grps'] = cepg_0
        quantities.append(cepg)

        # Death term
        a = 0.5*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.8*unit.percent/unit.year * np.ones(self.n_groups)
        deg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        self.ode_params['parole-death-rates-coeff'] = deg_0

        # Phase state
        self.population_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        self.population_phase.SetValue('feg', feg_0, self.initial_time)

        # Initialize inflows to zero
        self.ode_params['prison-inflow-rates'] = np.zeros(self.n_groups)

        return

    def run(self, *args):

        self.__zero_ode_parameters()

        time = self.initial_time

        while time <= self.end_time:

            # Interactions in the prison port
            #--------------------------------
            # two way "to" and "from" prison

            # to
            message_time = self.recv('prison')
            outflow_rates = self.__compute_outflow_rates( message_time, 'prison' )
            self.send( (message_time, outflow_rates), 'prison' )

            # from
            self.send( time, 'prison' )
            (check_time, prison_inflow_rates) = self.recv('prison')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['prison-inflow-rates'] = prison_inflow_rates

            # Interactions in the community port
            #-----------------------------------
            # one way "to" community

            message_time = self.recv('community')
            outflow_rates = self.__compute_outflow_rates( message_time, 'community' )
            self.send( (message_time, outflow_rates), 'community' )

            # Evolve parole group population to the next time stamp
            #------------------------------------------------------

            time = self.__step( time )

    def __rhs_fn(self, u_vec, t, params):

        feg = u_vec  # parole population groups

        prison_inflow_rates = params['prison-inflow-rates']

        inflow_rates  = prison_inflow_rates

        ce0g = self.ode_params['commit-to-community-coeff-grps']

        cepg = self.ode_params['commit-to-prison-coeff-grps']

        outflow_rates = ( ce0g + cepg ) * feg

        death_rates_coeff = params['parole-death-rates-coeff']

        death_rates = death_rates_coeff * feg

        dt_feg = inflow_rates - outflow_rates - death_rates

        return dt_feg

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        u_vec_0 = self.population_phase.GetValue('feg', time)
        t_interval_sec = np.linspace(0.0, self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                         u_vec_0, t_interval_sec,
                                         args=( self.ode_params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1,:]  # solution vector at final time step
        values = self.population_phase.GetRow(time) # values at previous time

        time += self.time_step

        self.population_phase.AddRow(time, values)

        # Update current values
        self.population_phase.SetValue('feg', u_vec, time)

        return time

    def __compute_outflow_rates(self, time, name):

      feg = self.population_phase.GetValue('feg',time)

      assert np.all(feg>=0.0), 'values: %r'%feg

      if name == 'prison':

          cepg = self.ode_params['commit-to-prison-coeff-grps']

          outflow_rates = cepg * feg

      if name == 'community':

          ce0g = self.ode_params['commit-to-community-coeff-grps']

          outflow_rates = ce0g * feg

      return outflow_rates

    def __zero_ode_parameters(self):
        '''
        If ports are not connected the corresponding outflows must be zero.

        '''

        zeros = np.zeros(self.n_groups)

        p_names = [p.name for p in self.ports]

        if 'community' not in p_names:
            self.ode_params['commit-to-community-coeff-grps']     = zeros

        if 'prison' not in p_names:
            self.ode_params['commit-to-prison-coeff-grps'] = zeros

        return



from cortix import Module
from cortix import Port
import numpy as np

class Body(Module):
    def __init__(self, mass=1, rad=(0, 0, 0), vel=(0, 0, 0), time=100, dt=0.01):
        super().__init__()

        self.G = 6.67408e-11
        self.ep = 1e-20

        self.mass = mass
        self.rad = rad

        self.vel = vel
        self.acc = np.zeros(3)

        self.other_bodies = None
        self.dt = dt
        self.time = time

        self.trajectory = []


    def force_from(self, other_mass, other_rad):
        delta = np.linalg.norm(other_rad - self.rad)
        if delta == 0:
            print("Collision!")
            exit(1)
        return (self.G * self.mass * other_mass) / (delta ** 2)

    def step(self):
        self.acc = np.zeros(3)
        total_force = np.zeros(3)

        for (body_mass, body_rad) in self.other_bodies:
            total_force += self.force_from(body_mass, body_rad)

        acc = total_force / self.mass
        self.acc = self.acc + acc
        self.vel = self.vel + acc * self.dt
        self.rad = self.rad + self.vel * self.dt

    def broadcast_data(self):
        # Broadcast (mass, pos) to every body
        for port in self.ports:
            self.send((self.mass, self.rad), port)

    def gather_data(self):
        # Get (mass, pos) from every other body
        self.other_bodies = [self.recv(port) for port in self.ports]

    def run(self):
        t = 0.0
        while t < self.time:
            self.broadcast_data()
            self.gather_data()
            self.step()
            self.trajectory.append(tuple(self.rad.flatten()))
            t += self.dt
            print(t)
        self.dump()

    def dump(self, file_name=None):
        if file_name is None:
            file_name = "body_{}.csv".format(id(self))
        with open(file_name, "w") as f:
            for (x, y, z) in self.trajectory:
                f.write("{}, {}, {}\n".format(x, y, z))

    def __repr__(self):
        return "{}".format([self.mass, self.rad, self.vel, self.acc])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
'''
Species class from Cortix support.
'''

import scipy.constants as const
from cortix.support.periodictable import ELEMENTS

class Species:
    """
    All SI units (kg,s,K,Pa,J,W).

    Notes
    -----
    The Species() class encapsulates either the molecular or empirical chemical formula of
    a compound. This is done as follows.

    Say `MAO2` is either a molecular or empirical chemical formula of a ficticious compound
    denoting minor actinides dioxide. The list of atoms is given as follows:

    ['0.49*Np-237', '0.42*Am-241', '0.08*Am-243', '0.01*Cm-244', '2.0*O-16']

    note the MA forming nuclides add to 1 = 0.49 + 0.42 + 0.08 + 0.01. Therefore
    the number of atoms in this compound is 3. 1 MA "ficticious" atom and 2 O.
    Note that the total number of "atoms" is obtained by summing all multipliers:
    0.49 + 0.42 + 0.08 + 0.01 + 2.0.
    The nuclide is indicated by the element symbol followed by a dash and the
    atomic mass number. Here the number of nuclide types is 5 (self.num_nuclide_types).

    The numbers preceeding the nuclide symbol before the * will be referred to as
    multipliers. The sum of the multipliers will add to the number of "atoms" in
    the formula. WARNING: a multiplier could be in the format 0.00e-00. In this
    case a hyphen may appear twice, e.g.: 1.549e-09*U-233

    Other forms can be used for common true species

    ['Np-237', '2.0*O-16'] or ['Np-237', 'O-16', 'O-16'] or [ '2*H', 'O' ] or
    [ 'H', 'O', 'H' ]  etc...

    This code will calculate the molar mass of any species with a given valid
    atom list using a provided periodic table of chemical elements. The user
    can also reset the value of the molar mass with a setter method.

    Attributes
    ----------

    """

    def __init__( self,
                  name='no-species-name',
                  formula_name='no-species-formula-name',
                  atoms=None,
                  charge=None,
                  flag='no-species-flag',
                  info=None,
                  latex_name='no-species-latex-name'):
        """Constructs a Species object.

        Parameters
        ----------
        name: str
            The name or "nick-name" of the species; often a short-hand name for the species,
            e.g. ndd for C12H26. The empirical formula name may be too long for typesetting into
            code, the name is usually a shorter variable name.

        formula_name: str
            The empirical formula name. See __atoms_from_formula_name() method for valid strings
            that allow for automatic build up of the list of atoms.

        atoms: list(str)
            List of atoms as described above in the class notes.

        flag: anytype
            Any data type specified by user. This can be used as an identification of the
            species when solving for concentration, or other designation of status.

        info: str
            A textual information about the species, e.g. the commercial name of the compound.

        latex_name: str
            A LaTeX string type varible typically generated automatically by this constructor.

        Examples
        --------

        """

        assert isinstance(name, str)
        self.name = name

        assert isinstance(formula_name, str)
        self.formula_name = formula_name

        if atoms is not None:
            assert isinstance(atoms, list)
            self.atoms = atoms
        else:
            self.atoms = list()

        if charge is not None:
            assert isinstance(charge, int)
            self.charge = charge
        else:
            self.charge = 0 # defaults to neutral species

        self.flag = flag  # flag can be any type

        self.info = info # info text such as technical name or other properties info

        assert isinstance(latex_name, str)
        self.latex_name = latex_name

        self.molar_mass = 0.0      # kg/mol
        self.molar_heat_pwr = 0.0
        self.molar_gamma_pwr = 0.0
        self.molar_radioactivity = 0.0

        self.molar_mass_unit = 'kg/mol'

        self.molar_heat_pwr_unit = 'W/mol'
        self.molar_gamma_pwr_unit = 'W/mol'
        self.molar_radioactivity_unit = 'Ci/mol'

        self.molar_radioactivity_fractions = list()

        if self.latex_name == 'no-species-latex-name' and self.formula_name != 'no-species-formula-name':
            self.__latex_name_from_formula_name()

        if len(self.atoms) == 0 and self.formula_name != 'no-species-formula-name':
            self.__atoms_from_formula_name()

        self.update_molar_mass()

    def update_molar_mass(self):
        """Recalculates the molar mass of the species and updates the species attribute.

        The typical use is to update the molar mass of the species after the list of atoms
        has been changed.

        """

        molar_mass_const = const.physical_constants['molar mass constant'][0]
        #molar_mass_const_unit = const.physical_constants['molar mass constant'][1]

        for entry in self.atoms:
            assert isinstance(entry, str)
            tmp = entry.split('*')
            nuclide = tmp[-1]
            element = nuclide.split('-')[0]
            if not element in ELEMENTS:
                print('Warning: not a chemical element %s'%element)
            #assert element in ELEMENTS, 'element = %r'%(element)

        self.num_atoms = 0
        self.num_nuclide_types = 0

        nuclides = dict()

        num_atoms = 0
        summ = 0.0
        for entry in self.atoms:
            assert isinstance(entry, str)
            # format example:  3.2*O-18, or 3*O or O or O-16
            tmp = entry.split('*')
            multiple = 1.0
            # single nuclide
            if len(tmp) == 1:
                nuclide = tmp[0]
            # multiple nuclide
            elif len(tmp) == 2:
                multiple = float(tmp[0])
                nuclide = tmp[1]
            else:
                assert False

            nuclides[nuclide] = multiple
            num_atoms += multiple

            try:
                tmp = nuclide.split('-')
                if len(tmp) == 1:
                    element = ELEMENTS[tmp[0]]
                    rel_atomic_mass = element.exactmass  # from isotopic composition
                    if rel_atomic_mass == 0.0:
                        rel_atomic_mass = element.mass
                    molar_mass = rel_atomic_mass * molar_mass_const
                elif len(tmp) == 2:
                    element = ELEMENTS[tmp[0]].isotopes[int(tmp[1].strip('m'))]
                    molar_mass = element.mass * molar_mass_const
                else:
                    assert False
            except KeyError:
                summ += multiple * 0.0
            else:
                summ += multiple * molar_mass

        self.molar_mass = summ

        self.num_atoms = num_atoms
        self.num_nuclide_types = len(nuclides)

        # Add charge
        if self.charge > 0:
            self.molar_mass -= self.charge * const.physical_constants['electron molar mass'][0]
        else:
            self.molar_mass += -1 * self.charge * const.physical_constants['electron molar mass'][0]

        # Exception: e^- (solvated electron)
        if self.formula_name == 'e^-':
            self.molar_mass = const.physical_constants['electron molar mass'][0]

    def ordered_atoms_list(self):
        """Sorted list of the atoms in the species; mostly for printing purposes.

        This method uses the internal list of atoms for a molecular or empirical formula and
        places it in order of decreasing magnitude of stoichiometric coefficient. For
        example, [O, 2*H] will be returned as [2*H, O]. This is used for
        printing purposes. The internal order will not change.

        Returns
        -------
        atoms2: list

        """

        atoms1 = self.atoms[:]  # shallow copy
        atoms2 = list()

        if len(self.atoms) <= 1:
            return atoms1

        if len(self.atoms) > 1:

            # save the multiplier value as a string type of scientific notation
            for entry in self.atoms:

                assert isinstance(entry, str)

                # format example:  3.2*O-18, or 3*O or O or O-16
                tmp = entry.split('*')

                nuclide = tmp[-1]

                multiplier = 0.0

                if len(tmp) == 1:
                    multiplier = 1
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                assert multiplier != 0.0, 'multiplier = %r' % (multiplier)

                multiplier = '{0:9.3e}'.format(multiplier)

                atoms1[self.atoms.index(entry)] = multiplier + '*' + nuclide

            # order in decreasing order of multiplier magnitude
            multipliers_lst = list()

            for entry in atoms1:

                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                multipliers_lst.append(float(multiplier))

            sorted_atoms = [a for (i, a) in sorted(zip(multipliers_lst, atoms1),
                key=lambda pair: pair[0], reverse=True)]

            atoms2 = sorted_atoms

        return atoms2

    def __atoms_from_formula_name(self):
        """Try to build the atoms list from the empirical formula.

        Examples that can be handled by the code below:

            *OH^-:              * for radical, ^ for charged species
            [CH2]3OH*^2-(a):    [] for groups of atoms
            H2O2(v):            () for phase

        This is useful for automating the creation of a species list of atoms, say from a
        reaction mechanism.

        """

        formula_name = self.formula_name.strip()

        # remove radical symbol *
        formula_name = formula_name.strip('*')
        tmp = formula_name.split('*')
        if len(tmp) > 1:
            assert len(tmp) == 2, 'fatal: only one "*" allowed.'
            formula_name = tmp[0]+tmp[1]

        # remove phase indicator
        i = formula_name.find('(')
        j = formula_name.rfind(')')
        if i != -1 and j != -1:
            formula_name = formula_name.replace(formula_name[i:j+1], '')
        elif (i == -1 and j != -1) or (i != -1 and j == -1):
            assert False, 'fatal: missing pairing ")".'

        # remove charge
        i = formula_name.find('^')
        if i != -1: # if success
            if formula_name[i+1].isnumeric():    # integer followed by sign
                charge = formula_name[i+1:i+3]
                formula_name = formula_name.replace(formula_name[i:i+3], '')
            else: # just a sign
                charge = formula_name[i+1:i+2]
                formula_name = formula_name.replace(formula_name[i:i+2], '')

            sign = charge[-1]

            if len(charge) == 2:
                val = int(charge[0])
            elif len(charge) == 1:
                val = 1
            else:
                assert False, 'fatal: invalid charge = %r'%charge

            if sign == '-':
                self.charge = -1 * val
            else:
                self.charge = val

        # find atom group multiplicity
        i = formula_name.find('[')
        if i != -1: # if success
            assert False, 'fatal: "[]" not yet implemented.' # TODO

        # build the atom list
        assert isinstance(self.atoms, list)
        assert len(self.atoms) == 0

        assert formula_name.isalnum(), 'fatal: formula name = %r'%formula_name

        # Find the chemical element symbol in the general form: Xy, say He, Na, O, etc.
        upper_case_ids = list()
        lower_case_ids = list()
        number_of_atoms_ids = list()

        for id,s in enumerate(formula_name):
            if s.isupper():
                upper_case_ids.append(id)
            if s.islower():
                lower_case_ids.append(id)
            if s.isnumeric():
                number_of_atoms_ids.append(id)

        for i in upper_case_ids:
            num_atoms = 1
            symbol = formula_name[i]
            if i+1 in number_of_atoms_ids:
                num_atoms = formula_name[i+1]
            if i+1 in lower_case_ids:
                symbol += formula_name[i+1]
                if i+2 in number_of_atoms_ids:
                    num_atoms = formula_name[i+2]

            if num_atoms == 1:
                self.atoms.append(symbol)
            else:
                self.atoms.append(num_atoms+'*'+symbol)

    def __latex_name_from_formula_name(self):
        """Try to build a LaTeX name from the empirical formula.

        Examples that can be handled by the code below:

            *OH^-               * for radical, ^ for charged species
            [CH2]3OH*^2-(a)     [] for groups of atoms
            H2O2(v)             () for phase

        Limitation to formulas with a maximum numeric value of 9.
        Unfinished.

        This is useful for automating the creation of a species list of atoms, say from a
        reaction mechanism. Important: to copy and paste the result into a LaTeX document, first
        had Python `print` the string and then copy the string. To use in a Jupyter notebook with
        markdown cells, again, print using a code cell and copy the output into a markdown cell.
        To use under LaTex, either enclose in $$ or drop it into a LaTex environment.

        """

        latex_name = str()

        formula_name = self.formula_name.strip()

        open_parenthesis = False
        open_parenthesis_after_numeric = False

        for idx, c_i in enumerate(formula_name):

            if c_i == '(':
                open_parenthesis = True
                if formula_name[idx-1].isnumeric():
                    latex_name += r'_{\mathrm{(' # escape \
                    open_parenthesis_after_numeric = True
                else:
                    latex_name += r'_\mathrm{(' # escape \
                continue

            if c_i == ')':
                if open_parenthesis_after_numeric:
                    latex_name += ')}}}'
                    open_parenthesis_after_numeric = False
                else:
                    latex_name += ')}'
                open_parenthesis = False
                continue

            if not open_parenthesis:

                if c_i == '*':
                    latex_name += r'^\bullet' # escape \
                    continue

                if c_i == '^':
                    assert idx < len(formula_name)-1, 'Error on %r'%formula_name
                    if formula_name[idx+1].isnumeric(): # there must be a charge sign
                        latex_name += '^{' + formula_name[idx+1:idx+3] + '}'
                    else: # no numeric, hence just a charge sign
                        latex_name += '^' + formula_name[idx+1]
                    continue

                if c_i.isalpha(): # alphabetic
                    latex_name += r'\mathrm{' + c_i + '}'
                    continue

                if idx < len(formula_name)-1:
                    if c_i.isnumeric() and formula_name[idx-1].isalpha() and formula_name[idx+1] != '(':
                        latex_name += '_{' + c_i + '}'
                    elif c_i.isnumeric() and formula_name[idx-1].isalpha() and formula_name[idx+1] == '(':
                        latex_name += '_{' + c_i
                    continue

                if c_i.isnumeric(): # numeric
                    latex_name += '_{' + c_i + '}'
                    continue

                if c_i in ['[', ']']:
                    assert False, 'fatal: "[]" not implemented yet.' #TODO

            else: # phase token
                latex_name += c_i


        self.latex_name = latex_name

    def __str__(self):
        s = '\n\t ' + \
            '\n\t Species(): name=%s;' + \
            ' formula_name=%s;' + \
            '\n\t formula=%s;' + \
            '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + \
            '\n\t charge=%s;' + \
            '\n\t flag=%s;' + \
            '\n\t info=%s;' + \
            '\n\t latex_name=%s;' + \
            '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t molar heat pwr=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + \
            '\n\t individual atoms=%s;' + \
            '\n\t molar radioactivity fractions=%s'
        return s % (self.name,
                self.formula_name,
                self.ordered_atoms_list(),
                self.num_atoms, self.num_nuclide_types, self.molar_mass, \
                        self.molar_mass_unit,
                self.charge,
                self.flag,
                self.info,
                self.latex_name,
                self.molar_radioactivity, self.molar_radioactivity_unit,
                self.molar_heat_pwr, self.molar_heat_pwr_unit,
                self.molar_gamma_pwr, self.molar_gamma_pwr_unit,
                [i.split('*')[-1] for i in self.atoms],
                ['%9.3e' % i for i in self.molar_radioactivity_fractions])

    def __repr__(self):
        s = '\n\t Species(): name=%s;' + \
            ' formula_name=%s;' + \
            '\n\t formula=%s;' + \
            '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + \
            '\n\t charge=%s;' + \
            '\n\t flag=%s;' + \
            '\n\t latex_name=%s;' + \
            '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t molar heat pwr=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + \
            '\n\t individual atoms=%s;' + \
            '\n\t molar radioactivity fractions=%s'
        return s % (self.name,
                self.formula_name,
                self.ordered_atoms_list(),
                self.num_atoms, self.num_nuclide_types, self.molar_mass, \
                        self.molar_mass_unit,
                self.charge,
                self.flag,
                self.latex_name,
                self.molar_radioactivity, self.molar_radioactivity_unit,
                self.molar_heat_pwr, self.molar_heat_pwr_unit,
                self.molar_gamma_pwr, self.molar_gamma_pwr_unit,
                [i.split('*')[-1] for i in self.atoms],
                ['%9.3e' % i for i in self.molar_radioactivity_fractions])

if __name__ == '__main__':
    tbp_org = Species( name='TBP', formula_name='[C4H9O]_3PO(o)',
              atoms=['12*C','27*H','4*O','P'] )
    print(tbp_org)
    no3Minus_aqu = Species( name='NO3-', formula_name='NO3^-(a)',
                   atoms=['N','3*O'] )
    print(no3Minus_aqu)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
"""
Cortix Run File
"""
import scipy.constants as unit
import matplotlib.pyplot as plt

from cortix import Cortix
from cortix import Network

from reactor import BWR
from turbine import Turbine
from condenser import Condenser
from params import get_params

def main():
    """Balance of plant of a boiling water nuclear reactor.

    Attributes
    ----------
    end_time: float
        End of the flow time in SI unit.
    time_step: float
        Size of the time step between port communications in SI unit.
    use_mpi: bool
        If set to `True` use MPI otherwise use Python multiprocessing.

    """

    # Preamble

    end_time = 30 * unit.minute
    time_step = 30.0 # seconds
    show_time = (True, 5*unit.minute)

    use_mpi = False  # True for MPI; False for Python multiprocessing
    plot_results = True # True for enabling plotting section below
    params = get_params() # parameters for BoP BWR

    #*****************************************************************************
    # Define Cortix system

    # System top level
    plant = Cortix(use_mpi=use_mpi, splash=True)

    # Network
    plant_net = plant.network = Network()

    params['start-time'] = 0
    params['end-time'] = end_time

    #*****************************************************************************
    # Create reactor module
    reactor = BWR(params)

    reactor.name = 'BWR'
    reactor.save = True
    reactor.time_step = time_step
    reactor.end_time = end_time
    reactor.show_time = show_time

    # Add reactor module to network
    plant_net.module(reactor)

    #*****************************************************************************
    # Create turbine 1 module
    params['turbine_inlet_pressure'] = 2
    params['turbine_outlet_pressure'] = 0.005
    params['high_pressure_turbine'] = True

    #params_turbine = reactor.params
    #params_turbine.inlet_pressure = 2
    #params.turbine_outlet_pressure = 0.5

    turbine1 = Turbine(params)

    turbine1.name = 'High Pressure Turbine'
    turbine1.save = True
    turbine1.time_step = time_step
    turbine1.end_time = end_time

    # Add turbine 1 module to network
    plant_net.module(turbine1)

    #*****************************************************************************
    # Create condenser module
    params['steam flowrate'] = params['steam flowrate'] * 2

    condenser = Condenser(params)

    condenser.name = 'Condenser'
    condenser.save = True
    condenser.time_step = time_step
    condenser.end_time = end_time

    plant_net.module(condenser)

    #*****************************************************************************
    # Create the BoP network connectivity
    plant_net.connect([reactor, 'coolant-outflow'], [turbine1, 'inflow'])
    plant_net.connect([turbine1, 'outflow-1'], [condenser, 'inflow-1'])
    plant_net.connect([condenser, 'outflow'], [reactor, 'coolant-inflow'])

    plant_net.draw()

    #*****************************************************************************
    # Run network dynamics simulation
    plant.run()

    #*****************************************************************************
    # Plot results

    if plot_results and (plant.use_multiprocessing or plant.rank == 0):

        # Reactor plots
        reactor = plant_net.modules[0]

        (quant, time_unit) = reactor.neutron_phase.get_quantity_history('neutron-dens')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']')

        plt.grid()
        plt.savefig('test-neutron-dens.png', dpi=300)

        (quant, time_unit) = reactor.neutron_phase.get_quantity_history('delayed-neutrons-cc')
        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']')

        plt.grid()
        plt.savefig('test-delayed-neutrons-cc.png', dpi=300)

        (quant, time_unit) = reactor.coolant_outflow_phase.get_quantity_history('temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']')

        plt.grid()
        plt.savefig('test-coolant-outflow-temp.png', dpi=300)

        (quant, time_unit) = reactor.reactor_phase.get_quantity_history('fuel-temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('test-fuel-temp.png', dpi=300)

        # Turbine1 plots
        turbine1 = plant_net.modules[1]

        (quant, time_unit) = turbine1.outflow_phase.get_quantity_history('power')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']',
                   title='High Pressure Turbine Power')
        plt.grid()
        plt.savefig('test-turbine1-power.png', dpi=300)

        (quant, time_unit) = turbine1.outflow_phase.get_quantity_history('temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']',
                   title='High Pressure Turbine Outflow Temperature')
        plt.grid()
        plt.savefig('test-turbine1-outflow-temp.png', dpi=300)

        # Condenser graphs
        condenser = plant_net.modules[-1]

        (quant, time_unit) = condenser.outflow_phase.get_quantity_history('temp')

        quant.plot(x_scaling=1/unit.minute, x_label='Time [m]',
                   y_label=quant.latex_name+' ['+quant.unit+']')
        plt.grid()
        plt.savefig('test-condenser-outflow-temp.png', dpi=300)

    # Properly shutdown simulation
    plant.close()
if __name__ == '__main__':
    main()



import os, time, datetime, threading, random, sys, string
import numpy as np
from cortix import Module
from cortix import Network
from cortix import Cortix
import matplotlib.pyplot as plt
import matplotlib

import matplotlib.image as mpimg
import matplotlib.animation as animation
import pandas as pd

class Particle_Plot(Module):
    def __init__(self, shape,modules=5,runtime=None, balls=1):
        super().__init__()
        self.filetime = str(datetime.datetime.now())[:10]
        self.fps = 60
        self.length = modules
        self.shape = shape
        self.timestamp=str(datetime.datetime.now())
        self.bndry = []
        self.colordic = dict()
        
    def run(self):
        print('start plot')
        self.dic = {}
        c = 0
        
        writer = animation.FFMpegWriter(fps=self.fps)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        x,y = [f[0] for f in self.shape],[f[1] for f in self.shape]
        ax.plot(x,y,'black')
        ax.axis('off')
        ax.autoscale()
        ax.set_aspect( 'equal', adjustable='datalim')
        ax.relim()
        
        modcount,self.oe=0,0
        self.linedic = {}
        tempdir='.writer'
        self.dpi = 90
        anifilename = 'particle_animation.mp4'.format(self.filetime)
        writer.setup(fig,anifilename, dpi=self.dpi)
        self.tke=0
        
        self.ke,self.elapsed = [],[]
        self.t_collisions = []
        oldtke=0
        while True:
            for i in self.ports:
                if not 'plot' in str(i):
                    continue
                lis = self.recv(i)
                self.send('hi',i)
                if isinstance(lis,str):
                    c+=1
                    self.color=lis
                    if c >=self.length:
                        print("Creating output figures")
                        writer.finish()
                        fig = plt.figure()
                        plt.plot(self.elapsed,self.ke)
                        plt.title("Total Kinetic Energy vs Time")
                        plt.savefig('energy_vs_time.png'.format(self.filetime),dpi=180)
##                        fig = plt.figure()
##                        plt.plot(self.elapsed,self.t_collisions)
##                        plt.title("Collisions with wall vs time")
##                        plt.savefig('collisions_vs_time.png'.format(self.filetime))
                        return
                    continue
##                collisions = 0
                for line in lis:
                    self.colordic[line.name] = [line.color, line.r]
                    if line.name not in self.dic:
                        self.dic[line.name]=[]
                        self.linedic[line.name], = ax.plot([],[],self.colordic[line.name][0])
                    self.tke += line.ke
##                    collisions += line.total_collisions
                    def xy(r,phi):
                        return [r*np.cos(phi), r*np.sin(phi)]
                    ucircle = xy(self.colordic[line.name][1],np.arange(0,6.28,0.01))
                    x,y = line.p[0]+ucircle[0],line.p[1]+ucircle[1]
                    self.linedic[line.name].set_data(x,y)
                modcount+=1
                elapsed = line.elapsed
                if round(elapsed,1) > self.oe:
                    print('Elapsed Time:', round(elapsed,1))
                    self.oe=round(elapsed,1)
                if modcount >= self.length:
                    self.ke.append(self.tke)
##                    if self.tke!=oldtke and abs(self.tke-oldtke)>1:
##                        print("Total KE:", self.tke,'Difference:', self.tke-oldtke)
##                        oldtke=self.tke
##                    self.t_collisions.append(collisions)
                    self.elapsed.append(elapsed)
                    self.tke=0
                    writer.grab_frame()
                    modcount = 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org
'''
Models a condenser with steam flowing perpendicular to several horizontal tubes.
Returns the runoff temperature of steam from the condenser.
'''

import logging
import math
import numpy as np
import matplotlib.pyplot as pyplot

import scipy.constants as unit
import iapws.iapws97 as steam_table

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class Condenser(Module):
    """Boiling water reactor single-point reactor.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `turbine`, and `pump`.
    See instance attribute `port_names_expected`.
    """

    # Class units
    meter = unit.meter = 1.0
    second = unit.second = 1.0

    m_per_s = unit.meter/unit.second
    kg_per_s = unit.kilo*unit.gram/unit.second

    def __init__(self):
        """Constructor.

        Parameters
        ----------
        None

        Notes
        -----

        VFdA: as of now both turbines are supposed to send the same information
              to the condenser. This is to be changed in the future to handle
              different turbine steam conditions outflow.

        Ports:

         + `inflow-1`: individual inflow from a turbine module.
         + `inflow-2`: individual inflow from a turbine module.
         + `outflow`: combined outflow from condenser.

        Attributes:

         + inflow_state: dict
            ['quality']: 0-1 quality of steam; 1 dew point steam;
                         >1 super heated; 0 bubble boint; <0 subcooled liquid.
            ['temp']: Runoff temperature from the turbine. Indicates how much
                      superheat must be removed before condensation can begin.
            ['pressure']: Not used.
            ['flowrate']: Half the amount of liquid (kg/s) that the condenser
                          must process each second.
            ['time']:     Current simulation time in seconds.

        """

        super().__init__()

        self.port_names_expected = ['inflow-1', 'inflow-2', 'outflow']

        # Units
        self.meter = Condenser.meter
        self.second = Condenser.second
        self.m_per_s = Condenser.m_per_s
        self.kg_per_s = Condenser.kg_per_s

        # General attributes
        self.initial_time = 0.0*self.second
        self.end_time = 0.0*self.second

        self.show_time = (False, 10.0)
        self.time_step = 10.0
        self.log = logging.getLogger('cortix')

        # Domain attributes
        self.inflow_state = None
        self.pipe_diameter = 0.1*self.meter
        self.liquid_velocity = 10.0*self.m_per_s
        self.cooling_water_flowrate = 100000.0*self.kg_per_s
        self.heat_transfer_area = 10200.0 # m2, or 500 4m long, 0.1m diameter pipes
        self.condensation_ht_coeff = 5000.0 # w/m-k
        self.subcooling_ht_coeff = 1000.0 # w/m-k
        self.ss_temp = 287.15 # Initial and ending temperature of the condenser
        self.inlet_pressure = 0.005 #MPa, the runoff pressure from the LPT
        self.coolant_inflow_temp = 287.15 #K, condenser coolant inflow temp
        # Outflow phase history
        quantities = list()

        flowrate = Quantity(name='condenser-runoff-flowrate',
                            formal_name='Condenser Runoff Flowrate',
                            unit='kg/s', value=0.0,
                            info='Condenser Outflow Flowrate',
                            latex_name=r'$Q$')

        quantities.append(flowrate)

        temp = Quantity(name='temp',
                        formal_name='Condenser Runoff Temp.',
                        unit='K',
                        value=0.0,
                        info='Condenser Outflow Temperature',
                        latex_name=r'$T_o$')

        quantities.append(temp)

        self.outflow_phase = Phase(time_stamp=self.initial_time, time_unit='s',
                                   quantities=quantities)

    def run(self, *args):

        time = self.initial_time
        end_time = self.end_time

        while time < end_time + self.time_step:

            if self.show_time[0] and abs(time%self.show_time[1]-0.0) <= 1.e-1:

                msg = 'time = '+str(round(time/unit.minute, 1))
                self.log.info(msg)

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------
            time = self.__step(time)

    def tester(self):
        """Passes a string of test data (temperature, chi) to the condenser.
           Debug the model itself.
        """
        x = 0
        #start with a basic temperature scale
        print(x, '\n', 'TESTING TEMPERATURE', '\n', x)
        temp = np.linspace(300, 500, 200)
        test_chi = 0
        output_vector = []
        for i in temp:
            output = self.__condenser(0, i, test_chi)
            print('temp is ', i, ' and outflow temp is ', output)
            output_vector.append(output)

        #test how the system responds to different chi's
        print(x, '\n', 'TESTING CHI', '\n', x)
        chi = np.linspace(0, 0.98, 100)
        test_temperature = steam_table._TSat_P(0.005)
        output_vector_2 = []
        for i in chi:
            output = self.__condenser(0, test_temperature, i)
            print('chi is ', i, ' and outflow temp is ', output)
            output_vector_2.append(output)

        #graph the results
        pyplot.plot(temp, output_vector)
        pyplot.title('inflow temperature vs outflow temperature')
        pyplot.savefig('condenser-inflow-temp-vs-outflow-temp.png')
        pyplot.close()
        pyplot.plot(chi, output_vector_2)
        pyplot.title('inflow quality vs outflow temperature')
        pyplot.savefig('inflow-quality-vs-outflow-temp.png')
        return

    def __call_ports(self, time):

        # Interactions in the inflow port
        #----------------------------------------
        # one way "from" inflow-1

        # from
        if self.get_port('inflow-1').connected_port:
            self.send(time, 'inflow-1')
            (check_time, inflow_state) = self.recv('inflow-1')
            assert abs(check_time-time) <= 1e-6

            self.inflow_state = inflow_state
            self.inflow_state['time'] = time

        # Interactions in the inflow port
        #----------------------------------------
        # one way "from" inflow-2

        # from
        if self.get_port('inflow-2').connected_port:
            self.send(time, 'inflow-2')
            (check_time, inflow_state) = self.recv('inflow-2')
            assert abs(check_time-time) <= 1e-6

            self.inflow_state = inflow_state
            self.inflow_state['time'] = time

        # Interactions in the outflow port
        #-----------------------------------------
        # one way "to" outflow

        if self.get_port('outflow').connected_port:
            message_time = self.recv('outflow')
            outflow_cool_temp = self.outflow_phase.get_value('temp', time)
            self.send((message_time, outflow_cool_temp), 'outflow')

    def __step(self, time):

        assert abs(time-self.inflow_state['time']) <= 1e-6

        temp_in = self.inflow_state['temp']
        chi_in = self.inflow_state['quality']
        flowrate = self.inflow_state['flowrate']

        t_out = self.__condenser(temp_in, chi_in, flowrate)

        condenser_runoff = self.outflow_phase.get_row(time)

        time = time + self.time_step

        self.outflow_phase.add_row(time, condenser_runoff)
        self.outflow_phase.set_value('temp', t_out)

        return time

    def __condenser(self, temp_in, chi_in, flowrate):
        """Simple model to condense a vapor-liquid mixture and subcool it.

        Takes in either superheated steam or a two-phase water mixture, and
        calculates the amount of surface area within a simple condenser required to
        remove all degrees of superheat and condense the mixture.
        """

        if flowrate == 0:
            # no flowrate, return initial simulation condition
            return self.ss_temp

        #temporary, basic model for condensation, until I can work out a more
        #advanced model that gives realistic values.
        pressure = self.inlet_pressure
        t_coolant = self.coolant_inflow_temp
        h_mixture = steam_table._Region4(pressure, chi_in)['h']
        h_dewpt = steam_table._Region4(pressure, 0)['h']
        heat_pwr = (h_mixture - h_dewpt) * unit.kilo * flowrate
        area_required = heat_pwr/(self.condensation_ht_coeff * (temp_in - t_coolant))
        area_remaining = self.heat_transfer_area - area_required

        if area_remaining > 0:
            initial_coolant_temp = self.coolant_inflow_temp
            delta_t = heat_pwr/(self.cooling_water_flowrate * unit.kilo * 4.184)
            final_coolant_temp = initial_coolant_temp + delta_t
            subcooling_pwr = area_remaining * self.subcooling_ht_coeff *\
                             (temp_in - final_coolant_temp)
            delta_tb = subcooling_pwr/(flowrate * 4.184 * unit.kilo)
            runoff_temp = temp_in - delta_tb

            if runoff_temp < initial_coolant_temp:
                runoff_temp = initial_coolant_temp

            return runoff_temp

        return self.ss_temp
        #anything below this point is not currently in use; I need to figure
        #out a more accurate condensation model, the current one calculates
        #heat transfer coefficients in excess of 100k w/m2-k

        #print(time, ' chi_in is ', chi_in)
        critical_temp = steam_table._TSat_P(self.inlet_pressure)
        condenser_runoff = self.coolant_inflow_temp
        #if chi_in == -1 and temp_in > critical_temp: #superheated vapor inlet; deprecated
            #return 293.15
            #x = x_in
            #pressure = 0.005
            #h_steam = steam_table._Region2(temp_in, pressure)['h']
            #h_sat = steam_table._Region4(0.005, 1)['h']
            #q = params['steam flowrate'] * (h_steam - h_sat) * sc.kilo

            #log mean temperature difference
            #delta_Ta = temp_in - critical_temp
            #delta_Tb = 20-10
            #LMTD = (delta_Ta - delta_Tb) / (math.log(delta_Ta) - math.log(delta_Tb))

            #low = steam_table.IAPWS97(T=critical_temp+1, P=pressure)
            #high = steam_table.IAPWS97(T=temp_in, P=pressure)

            #low_conductivity = low.k
            #low_rho = low.rho
            #low_prandtl = low.Prandt
            #low_viscosity = low.mu
            #low_reynolds = (low_rho * 0.1 * 40)/low_viscosity
            #low_nusselt = (0.3 + 0.629 * (low_reynolds ** 0.5) * low_prandtl**(1/3))
            #low_nusselt = low_nusselt/(1+(0.4/low_prandtl)**(2/3))**(1/4)
            #low_nusselt = low_nusselt * (1 + (low_reynolds / 282000)**(5/6))**(4/4.5)
            #nusselt number at the crtical temperature according to the
            #churchill-bernsetin correlation
            #low_heat_transfer_coeff = low_nusselt * (low.k/0.1)

            #high_rho = high.rho
            #high_prandtl = high.Prandt
            #high_viscocity = high.mu
            #high_conductivity = high.k
            #high_viscosity = high.mu
            #high_reynolds = (high_rho * 0.1 * 40)/high_viscosity
            #high_nusselt = (0.3 + 0.629 * (high_reynolds ** 0.5) * high_prandtl**(1/3))
            #high_nusselt = high_nusselt/(1+(0.4/high_prandtl)**(2/3))**(1/4)
            #high_nusselt = high_nusselt * (1 + (high_reynolds / 282000)**(5/6))**(4/4.5)
            #high_heat_transfer_coeff = high_nusselt * (high_conductivity/0.1)

            #average_ht_transfer_coeff = (high_heat_transfer_coeff + low_heat_transfer_coeff)/2

        if chi_in > 0:
            pressure = self.inlet_pressure
            h_mixture = steam_table._Region4(pressure, chi_in)['h']
            h_dewpt = steam_table._Region4(pressure, 0)['h']
            heat_pwr = (h_mixture - h_dewpt) * unit.kilo * flowrate

            critical_temp = temp_in
            critical_temp = steam_table._TSat_P(pressure)
            sat_liquid = steam_table.IAPWS97(T=critical_temp - 1, P=pressure)
            water_cp = steam_table.IAPWS97(T=287.15, P=0.1013).Liquid.cp
            delta_tb = heat_pwr/(self.cooling_water_flowrate * \
                    water_cp * unit.kilo)
            t_c_in = self.coolant_inflow_temp
            t_c_out = t_c_in + delta_tb

            lmtd = (t_c_out - t_c_in)/(math.log(critical_temp - t_c_in))/\
                    (critical_temp - t_c_out)

            pipe_diameter = self.pipe_diameter
            liquid_velocity = self.liquid_velocity

            #calculate the convective heat transfer coefficient on a liquid basis
            #from the Churchill-Bernstein correlation
            liquid_conductivity = sat_liquid.Liquid.k
            liquid_rho = sat_liquid.Liquid.rho
            liquid_prandtl = sat_liquid.Liquid.Prandt
            liquid_viscosity = sat_liquid.Liquid.mu
            liquid_reynolds = (liquid_rho * pipe_diameter * liquid_velocity)/\
                    liquid_viscosity

            liquid_nusselt = (0.3 + 0.629 * (liquid_reynolds ** 0.5) * \
                    liquid_prandtl**(1/3))
            liquid_nusselt = liquid_nusselt/(1+(0.4/liquid_prandtl)**(2/3))**(1/4)
            #nusselt number at the critical temperature according to the
            #churchill-bernstein correlation
            liquid_nusselt = liquid_nusselt * (1 + (liquid_reynolds / 282000)**(5/6))**(4/4.5)
            #overall convective heat transfer coefficient
            liquid_heat_transfer_coeff = liquid_nusselt * (liquid_conductivity/pipe_diameter)

            #determine the Martinelli paremeter
            steam = steam_table.IAPWS97(T=critical_temp+1, P=pressure)
            steam_rho = steam.Vapor.rho
            steam_viscosity = steam.Vapor.mu

            #print(chi_in)

            xtt = (((1 -chi_in)/chi_in)**1.8 * (steam_rho/liquid_rho) * (liquid_viscosity/
                                                                         steam_viscosity)**0.5)
            xtt = math.sqrt(xtt)

            #determine the overall heat transfer coefficient from the McNaught
            #expression
            alpha_sh = liquid_heat_transfer_coeff * (1/xtt)**0.78
            print(alpha_sh, 'alpha_sh')

            #determine how far the two-phase mixture gets in the condenser
            #before being fully condensed
            condensation_area = (heat_pwr/(alpha_sh * lmtd))

            remaining_area = self.heat_transfer_area - condensation_area
            condenser_runoff = critical_temp
            print(remaining_area)
            #if time > params['malfunction start'] and time < params['malfunction end']:
                #condenser_runoff = 14 + 273.15

            if remaining_area < 0:
                condenser_runoff = critical_temp

            elif remaining_area > 0:
            #subcool the remaining liquid
            #iterative method
            #0. guess an ending value for the coolant and
            #1. calculate high nusselt number
            #2. guess a value for low temperature
            #3. calculate low nusselt number
            #4. calculate lmtd and average heat transfer coefficient
            #5. calculate q
            #6. use q to calculate a new ending temperature for the water and the
            #   vapor
            #7. rinse and repeat

                # initial guess
                final_coolant_temperature_guess = steam_table._TSat_P(self.inlet_pressure) - 10
                final_runoff_temperature_guess = 300 # initial guess
                final_runoff_temperature = 0 #iteration variable

                # Loop until convergeance, +- 1 degree kelvin
                # Compute the runoff temperature (subcooled liquid)
                while abs((final_runoff_temperature_guess -
                           final_runoff_temperature)) > 1:

                    runoff_in = steam_table._TSat_P(self.inlet_pressure)
                    coolant_in = self.coolant_inflow_temp #k
                    #assert(final_coolant_temperature_guess < final_runoff_temperature_guess)
                    #assert(final_coolant_temperature_guess > coolant_in)

                    final_runoff_temperature = final_runoff_temperature_guess

                    #guess the LMTD
                    lmtd = ((runoff_in - final_coolant_temperature_guess) - (final_runoff_temperature - coolant_in)) / math.log(((runoff_in -  final_coolant_temperature_guess) / (final_runoff_temperature - coolant_in)))

                    #average wall temperatures
                    upper_wall_temp = (coolant_in + runoff_in)/2
                    lower_wall_temp = (final_runoff_temperature + \
                            final_coolant_temperature_guess)/2

                    #high temperature heat transfer coefficient calculation
                    high_properties = steam_table.IAPWS97(T=upper_wall_temp,
                                                          P=pressure)
                    high_conductivity = high_properties.Liquid.k
                    high_rho = high_properties.Liquid.rho
                    high_prandtl = high_properties.Liquid.Prandt
                    high_viscosity = high_properties.Liquid.mu
                    high_reynolds = (high_rho * pipe_diameter * liquid_velocity)/\
                            high_viscosity

                    high_nusselt = (0.3 + 0.629 * (high_reynolds ** 0.5) * \
                            high_prandtl**(1/3))
                    high_nusselt = high_nusselt/(1+(0.4/high_prandtl)**(2/3))**(1/4)
                    high_nusselt = high_nusselt * (1 + (high_reynolds / 282000)**(5/6))**(4/4.5)
                    #overall convective heat transfer coefficient
                    high_heat_transfer_coeff = high_nusselt * (high_conductivity/pipe_diameter)

                    #   low temperature heat transfer coefficient calculation
                    low_properties = steam_table.IAPWS97(T=lower_wall_temp, P=pressure)
                    low_conductivity = low_properties.Liquid.k
                    low_rho = low_properties.Liquid.rho
                    low_prandtl = low_properties.Liquid.Prandt
                    low_viscosity = low_properties.Liquid.mu
                    low_reynolds = (low_rho * pipe_diameter * \
                            liquid_velocity)/low_viscosity

                    low_nusselt = 0.3 + 0.629 * (low_reynolds ** 0.5) * \
                            low_prandtl**(1/3)
                    low_nusselt = low_nusselt/(1+(0.4/low_prandtl)**(2/3))**(1/4)
                    low_nusselt = low_nusselt * (1 + (low_reynolds / 282000)**(5/6))**(4/4.5)
                    #overall convective heat transfer coefficient
                    low_heat_transfer_coeff = low_nusselt * (low_conductivity/pipe_diameter)

                    #average heat transfer coefficient
                    average_heat_transfer_coeff = ((high_heat_transfer_coeff
                                                    + low_heat_transfer_coeff)/2)

                    #calculate q = UAdeltaTlm
                    q_1 = average_heat_transfer_coeff * remaining_area * lmtd/10

                    #calculate actual coolant and liquid water temperatures

                    coolant_cp = steam_table.IAPWS97(T=coolant_in, P=self.inlet_pressure).Liquid.cp
                    iterated_coolant_final_temp = ((q_1/(coolant_cp *
                                                         self.cooling_water_flowrate
                                                         * unit.kilo)) + coolant_in)

                    steam_cp = steam_table.IAPWS97(T=runoff_in, P=self.inlet_pressure).Vapor.cp
                    iterated_steam_final_temp = (runoff_in - (q_1/(steam_cp *
                                                                   flowrate *
                                                                   unit.kilo)))

                    if (iterated_steam_final_temp < coolant_in or
                            iterated_coolant_final_temp > runoff_in or
                            iterated_steam_final_temp < iterated_coolant_final_temp):

                        final_runoff_temperature_guess = coolant_in
                        #final_coolant_runoff_guess = runoff_in use for later implementation
                        break

                    final_runoff_temperature_guess = iterated_steam_final_temp
                    final_coolant_temperature_guess = iterated_coolant_final_temp

                # end of: Loop until convergeance, +- 1 degree kelvin
                condenser_runoff = final_runoff_temperature_guess

        else:
            condenser_runoff = self.ss_temp

        return condenser_runoff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from .src.cortix_main import Cortix

from .src.network import Network
from .src.module import Module
from .src.port import Port

from .support.phase import Phase
from .support.quantity import Quantity
from .support.species import Species
from .support.chemeng.reaction_mechanism import ReactionMechanism
from .support.chemeng.reaction_mechanism import print_reaction_sub_mechanisms



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org

from multiprocessing import Pipe

class Port:
    """Provides a method of communication between modules.

    The Port class provides an interface for creating ports and connecting them to
    other ports for the purpose of data transfer. Data exchange takes place by
    send and/or receive calls on a given port. The concept of a port is that of a data
    transfer "interaction." This can be one- or two-way with sends and receives.
    A port is connected to only one other port; as two ends of a pipe are connected.
    """

    def __init__(self, name=None, use_mpi=False):
        '''Constructs a Port object

        Parameters
        ----------
        name: str
          The name of the Port object
        use_mpi: bool
          True for MPI, False for Multiprocessing

        Attributes
        ----------
            id: int
            name: string
            use_mpi: bool

        '''

        self.id = None
        self.name = name
        self.use_mpi = use_mpi

        if self.use_mpi:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.rank = None
        else:
            self.pipe = None

        self.connected_port = None

    def connect(self, port):
        '''Connect this port to another port

        Ports must be connected for data to flow between them.

        Parameters
        ----------
        port: Port
           A Port object to connect to.

        Returns
        -------
        None

        '''

        assert isinstance(port, Port), 'Connecting port must be of Port type'

        self.connected_port = port
        port.connected_port = self
        port.use_mpi = self.use_mpi

        if not port.use_mpi:
            (self.pipe, port.pipe) = Pipe()

    def send(self, data, tag=None):
        '''
        Send data to the connected port.

        If the sending port is not connected do nothing.

        Parameters
        ----------
        data: any
           This data must be pickleable.
        tag: int, optional
           MPI tag used in sending data.

        '''

        if not tag:
            tag = self.id

        if self.connected_port:
            if self.use_mpi:
                # This is an MPI blocking send
                self.comm.send(data, dest=self.connected_port.rank, tag=tag)
            else:
                self.pipe.send(data)

        return

    def __is_connected(self):
        '''
        Check for a connected port.

        Parameters
        ----------
        None

        Returs
        ------
        False or True: boolean
        Default False.

        '''
        if self.connected_port:
            return True
        else:
            return False
    is_connected = property(__is_connected, None, None, None)

    def recv(self):
        '''
        Receive data from the connected port.

        Warning
        -------
        This function will block if no data has been sent yet.

        Returns
        --------
        data: any

        '''
        if self.connected_port:
            if self.use_mpi:
                # This is an MPI blocking receive
                return self.comm.recv(source=self.connected_port.rank,
                        tag=self.connected_port.id)
            else:
                return self.pipe.recv()

        return

    def __eq__(self, other):
        '''Check for port equality'''

        return self.name == other.name

    def __repr__(self):
        '''Port name representation'''

        return self.name

if __name__ == '__main__':
    # Create some ports
    p1 = Port('test1')
    p2 = Port('test2')

    # Connect the ports
    p1.connect(p2)

    # View connections
    print(p1)
    print(p2)
    print(p1.connected_port)
    print(p2.connected_port)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import numpy as np
import scipy.constants as unit
from scipy.integrate import odeint

from cortix import Module
from cortix import Phase
from cortix import Quantity

class Jail(Module):
    '''
    Jail Cortix module used to model criminal group population in a jail.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `probation`, `adjudication`, `arrested`, `prison`, and `community`.
    See instance attribute `port_names_expected`.

    '''

    def __init__(self, n_groups=1, pool_size=0.0):
        '''
        Parameters
        ----------
        n_groups: int
            Number of groups in the population.
        pool_size: float
            Upperbound on the range of the existing population groups. A random value
            from 0 to the upperbound value will be assigned to each group.

        '''

        super().__init__()

        self.port_names_expected = ['probation','adjudication','arrested','prison',
                                    'community']
        quantities      = list()
        self.ode_params = dict()

        self.initial_time = 0.0 * unit.day
        self.end_time     = 100 * unit.day
        self.time_step    = 0.5 * unit.day

        unit.month = 30*unit.day
        unit.percent = 1/100

        # Population groups
        self.n_groups = n_groups

        # Jail population groups
        fjg_0 = np.random.random(self.n_groups) * pool_size
        fjg = Quantity(name='fjg', formal_name='jail-pop-grps',
                latex_name = '$n_j^{(g)}$',
                unit='# offenders', value=fjg_0, info='Jail Population Groups')
        quantities.append(fjg)

        # Model parameters: commitment coefficients

        # Jail to community
        a =  35*unit.percent/unit.year * np.ones(self.n_groups)
        b =  40*unit.percent/unit.year * np.ones(self.n_groups)
        cj0g_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        cj0g = Quantity(name='cj0g', formal_name='commit-community-coeff-grps',
               unit='1/s', value=cj0g_0)
        self.ode_params['commit-to-community-coeff-grps'] = cj0g_0
        quantities.append(cj0g)

        # Jail to prison    
        a =  60*unit.percent/unit.year * np.ones(self.n_groups)
        b =  65*unit.percent/unit.year * np.ones(self.n_groups)
        cjpg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        cjpg = Quantity(name='cjpg', formal_name='commit-prison-coeff-grps',
               unit='1/s', value=cjpg_0)
        self.ode_params['commit-to-prison-coeff-grps'] = cjpg_0
        quantities.append(cjpg)

        # Death term
        a = 0.5*unit.percent/unit.year * np.ones(self.n_groups)
        b = 0.6*unit.percent/unit.year * np.ones(self.n_groups)
        djg_0 = (a + (b-a)*np.random.random(self.n_groups)) / self.n_groups
        self.ode_params['jail-death-rates-coeff'] = djg_0

        # Phase state
        self.population_phase = Phase(self.initial_time, time_unit='s',
                quantities=quantities)

        self.population_phase.SetValue('fjg', fjg_0, self.initial_time)

        # Initialize inflows to zero
        self.ode_params['arrested-inflow-rates']     = np.zeros(self.n_groups)
        self.ode_params['probation-inflow-rates']    = np.zeros(self.n_groups)
        self.ode_params['adjudication-inflow-rates'] = np.zeros(self.n_groups)

        return

    def run(self, *args):

        self.__zero_ode_parameters()

        time = self.initial_time

        while time <= self.end_time:

            # Interactions in the prison port
            #--------------------------------
            # one way "to" prison

            message_time = self.recv('prison')
            outflow_rates = self.__compute_outflow_rates( message_time, 'prison' )
            self.send( (message_time, outflow_rates), 'prison' )

            # Interactions in the adjudication port
            #------------------------------------
            # one way "from" adjudication

            self.send( time, 'adjudication' )
            (check_time, adjudication_inflow_rates) = self.recv('adjudication')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['adjudication-inflow-rates'] = adjudication_inflow_rates

            # Interactions in the arrested port
            #----------------------------------
            # one way "from" arrested

            self.send( time, 'arrested' )
            (check_time, arrested_inflow_rates) = self.recv('arrested')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['arrested-inflow-rates'] = arrested_inflow_rates

            # Interactions in the probation port
            #-----------------------------------
            # one way "from" probation

            self.send( time, 'probation' )
            (check_time, probation_inflow_rates) = self.recv('probation')
            assert abs(check_time-time) <= 1e-6
            self.ode_params['probation-inflow-rates'] = probation_inflow_rates

            # Interactions in the community port
            #------------------------------
            # one way "to" community

            message_time = self.recv('community')
            outflow_rates = self.__compute_outflow_rates( message_time, 'community' )
            self.send( (message_time, outflow_rates), 'community' )

            # Evolve jail group population to the next time stamp
            #----------------------------------------------------

            time = self.__step( time )

    def __rhs_fn(self, u_vec, t, params):

        fjg = u_vec  # jail population groups

        arrested_inflow_rates    = params['arrested-inflow-rates']
        probation_inflow_rates   = params['probation-inflow-rates']
        adjudication_inflow_rates = params['adjudication-inflow-rates']

        inflow_rates  = arrested_inflow_rates + probation_inflow_rates + \
                        adjudication_inflow_rates

        cj0g = self.ode_params['commit-to-community-coeff-grps']

        cjpg = self.ode_params['commit-to-prison-coeff-grps']

        outflow_rates = ( cj0g + cjpg ) * fjg

        death_rates_coeff = params['jail-death-rates-coeff']

        death_rates = death_rates_coeff * fjg

        dt_fjg = inflow_rates - outflow_rates - death_rates

        return dt_fjg

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`u = (u_1(0),u_2(0),\ldots)`
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None

        '''

        u_vec_0 = self.population_phase.GetValue('fjg', time)
        t_interval_sec = np.linspace(0.0, self.time_step, num=2)

        (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                         u_vec_0, t_interval_sec,
                                         args=( self.ode_params, ),
                                         rtol=1e-4, atol=1e-8, mxstep=200,
                                         full_output=True)

        assert info_dict['message'] =='Integration successful.', info_dict['message']

        u_vec = u_vec_hist[1,:]  # solution vector at final time step
        values = self.population_phase.GetRow(time) # values at previous time

        time += self.time_step

        self.population_phase.AddRow(time, values)

        # Update current values
        self.population_phase.SetValue('fjg', u_vec, time)

        return time

    def __compute_outflow_rates(self, time, name):

      fjg = self.population_phase.GetValue('fjg',time)

      assert np.all(fjg>=0.0), 'values: %r'%fjg

      if name == 'prison':

          cjpg = self.ode_params['commit-to-prison-coeff-grps']

          outflow_rates = cjpg * fjg

      if name == 'community':

          cj0g = self.ode_params['commit-to-community-coeff-grps']

          outflow_rates = cj0g * fjg

      return outflow_rates

    def __zero_ode_parameters(self):
        '''
        If ports are not connected the corresponding outflows must be zero.

        '''

        zeros = np.zeros(self.n_groups)

        p_names = [p.name for p in self.ports]

        if 'community' not in p_names:
            self.ode_params['commit-to-community-coeff-grps']     = zeros

        if 'prison' not in p_names:
            self.ode_params['commit-to-prison-coeff-grps'] = zeros

        return
